import inspect
from collections.abc import Sequence
from typing import Any

import pandas as pd
from pandas.io.formats.style import Styler

from pepbench.utils._rename_maps import (
    _algo_level_mapping,
    _algorithm_mapping,
    _metric_mapping,
    _nan_reason_mapping,
    _nan_reason_mapping_short,
)

__all__ = [
    "convert_to_latex",
    "create_algorithm_result_table",
    "create_nan_reason_table",
    "create_outlier_correction_table",
    "create_reference_pep_table",
]


def create_reference_pep_table(data: pd.DataFrame) -> pd.DataFrame:
    # Create a new DataFrame with formatted columns
    formatted_data = data.copy()
    formatted_data.index.names = [s.capitalize() for s in formatted_data.index.names]

    formatted_data[("pep_ms", "mean_std")] = formatted_data.apply(
        lambda row: f"{row[('pep_ms', 'mean')]:.2f}({row[('pep_ms', 'std')]:.2f})",
        axis=1,
    )
    formatted_data[("pep_ms", "range")] = formatted_data.apply(
        lambda row: f"[{int(row[('pep_ms', 'min')])}, {int(row[('pep_ms', 'max')])}]",
        axis=1,
    )

    # Select only the required columns for the LaTeX table
    formatted_data = formatted_data[[("pep_ms", "mean_std"), ("pep_ms", "range")]]
    formatted_data.columns = ["M ± SD [ms]", "Range [ms]"]

    return formatted_data


def create_algorithm_result_table(data: pd.DataFrame, collapse_algo_levels: bool = False) -> pd.DataFrame:
    data = data.copy()

    # Create a new DataFrame with formatted columns
    formatted_data = pd.DataFrame(index=data.index)
    algo_levels = [_algo_level_mapping[s] for s in data.index.names if s in _algo_level_mapping]
    formatted_data.index.names = [_algo_level_mapping[s] for s in formatted_data.index.names]

    export_cols_m_sd = ["Mean Absolute Error [ms]", "Mean Error [ms]", "Mean Absolute Relative Error [%]"]
    export_cols_m_sd = [s for s in export_cols_m_sd if s in data.columns]

    for key in export_cols_m_sd:
        # formatted_data[key] = data.apply(lambda row: print(row[(key, "mean")]), axis=1)
        formatted_data[key] = data.apply(
            lambda row, k=key: rf"{row[(k, 'mean')]:.1f} \pm {row[(k, 'std')]:.1f}", axis=1
        )

    if "Invalid PEPs" in data.columns:
        formatted_data[r"Invalid PEPs"] = data.apply(
            lambda row: f"{int(row[('Invalid PEPs', 'total')])} "
            rf"({(row[('Invalid PEPs', 'total')] / row[('Total PEPs', 'total')] * 100):.1f} \%)",
            axis=1,
        )

    # rename index values
    formatted_data = formatted_data.rename(index=_algorithm_mapping)

    if collapse_algo_levels:
        formatted_data = formatted_data.reset_index()
        formatted_data["algo_merged"] = formatted_data.apply(lambda row: " | ".join(row[algo_levels]), axis=1)
        formatted_data = formatted_data.set_index("algo_merged").drop(columns=algo_levels)

    return formatted_data


def create_outlier_correction_table(data: pd.DataFrame, outlier_algos: Sequence[str] | None = None) -> pd.DataFrame:
    data_index = data.index.get_level_values("b_point_algorithm").unique()
    data = data.groupby("b_point_algorithm", group_keys=False).apply(
        lambda df: df.reindex(outlier_algos, level="outlier_correction_algorithm")
    )
    data = data.reindex(data_index, level="b_point_algorithm")
    data = data.rename(index=_algorithm_mapping)
    data = data.rename(_metric_mapping, axis=1)
    data.index.names = [_algo_level_mapping[s] for s in data.index.names]
    return data


def create_nan_reason_table(
    data: pd.DataFrame, outlier_algos: Sequence[str] | None = None, use_short_names: bool = True
) -> pd.DataFrame:
    data = data[[("nan_reason", "estimated")]].dropna().droplevel(level=-1, axis=1)
    data = data.groupby(["b_point_algorithm", "outlier_correction_algorithm"]).value_counts()
    data = data.unstack().fillna(0).astype("Int64")
    if outlier_algos is not None:
        data = data.reindex(outlier_algos, level="outlier_correction_algorithm")
    data.columns.name = "Reason"
    if use_short_names:
        data = data.rename(columns=_nan_reason_mapping_short)
    else:
        data = data.rename(columns=_nan_reason_mapping)
    data = data.rename(index=_algorithm_mapping)
    data.index.names = [_algo_level_mapping[s] for s in data.index.names]

    return data


def convert_to_latex(
    data: pd.DataFrame | Styler, collapse_index_columns: bool = False, **kwargs: dict[str, Any]
) -> str:
    kwargs.setdefault("hrules", True)
    kwargs.setdefault("position", "ht")
    kwargs.setdefault("siunitx", True)
    kwargs.setdefault("clines", "skip-last;data")
    if "environment" not in kwargs:
        kwargs.setdefault("position_float", "centering")
    kwargs.setdefault("convert_css", True)
    rename_map = {
        "Q-Peak Detection": "Algorithm",
        "Mean Absolute Error [ms]": r"\ac{MAE} [ms]",
        "Mean Error [ms]": r"\ac{ME} [ms]",
        "Mean Absolute Relative Error [%]": r"\ac{MARE} [%]",
        "Invalid PEPs": r"Invalid\newline PEPs",
    }

    if isinstance(data, pd.DataFrame):
        if collapse_index_columns:
            data = data.reset_index()

        data = data.rename(columns=rename_map)
        styler = data.style
    if isinstance(data, Styler):
        styler = data

    if collapse_index_columns:
        styler = styler.hide(axis=0)

    styler = styler.format_index(lambda x: x.replace("%", r"\%"), axis=1)
    if kwargs.get("column_header_bold", False):
        styler = styler.map_index(lambda _: "font-weight: bold;", axis=1)
    if kwargs.get("row_header_bold", False):
        styler = styler.map_index(lambda _: "font-weight: bold;", axis=0)

    if kwargs.get("escape_index", False):
        styler = styler.format_index(escape="latex", axis=0)
    if kwargs.get("escape_columns", False):
        styler = styler.format_index(escape="latex", axis=1)

    kwargs_styler = {k: v for k, v in kwargs.items() if k in list(inspect.signature(styler.to_latex).parameters.keys())}

    latex_str = styler.to_latex(**kwargs_styler)

    latex_str = latex_str.replace(
        "\\centering",
        "\\centering\n\\sisetup{separate-uncertainty=true,multi-part-units=single,table-align-uncertainty=true}",
    )
    return latex_str



"""Individual algorithms for preprocessing signals and extracting fiducial points from ECG and ICG signals.

Together, these algorithms can be combined into a PEP extraction pipeline
(e.g., :class:`~pepbench.pipelines.PepExtractionPipeline`).

The following categories of algorithms are available:
    * :mod:`~pepbench.algorithms.heartbeat_segmentation` - Algorithms for heartbeat segmentation
    * :mod:`~pepbench.algorithms.preprocessing` - Algorithms for preprocessing of ECG and ICG signals
    * :mod:`~pepbench.algorithms.ecg` - Algorithms for ECG fiducial point extraction
    * :mod:`~pepbench.algorithms.icg` - Algorithms for ICG fiducial point extraction
    * :mod:`~pepbench.algorithms.outlier_correction` - Algorithms for outlier correction of ICG fiducial points

"""

from pepbench.algorithms import ecg, heartbeat_segmentation, icg, outlier_correction, preprocessing

__all__ = ["ecg", "heartbeat_segmentation", "icg", "outlier_correction", "preprocessing"]



"""Module for plotting PEP benchmark results."""

import inspect
from collections.abc import Callable, Sequence
from typing import Any

import biopsykit as bp
import numpy as np
import pandas as pd
import pingouin as pg
import seaborn as sns
from fau_colors.v2021 import cmaps, colors_all
from matplotlib import pyplot as plt

from pepbench.data_handling import get_data_for_algo
from pepbench.data_handling._data_handling import get_performance_metric, get_reference_data
from pepbench.plotting._base_plotting import _plot_blandaltman, _plot_paired
from pepbench.plotting._utils import _get_fig_ax, _get_fig_axs, _remove_duplicate_legend_entries
from pepbench.utils._rename_maps import (
    _algo_level_mapping,
    _algorithm_mapping,
    _metric_mapping,
    _xlabel_mapping,
    _ylabel_mapping,
)

__all__ = [
    "boxplot_algorithm_performance",
    "boxplot_reference_pep",
    "histplot_heart_rate",
    "paired_plot_error_outlier_correction",
    "paired_plot_error_pep_pipeline",
    "plot_q_wave_detection_waveform_detailed_comparison",
    "regplot_error_bmi",
    "regplot_error_heart_rate",
    "regplot_pep_heart_rate",
    "residual_plot_pep",
    "residual_plot_pep_bmi",
    "residual_plot_pep_heart_rate",
    "residual_plot_pep_participant",
    "residual_plot_pep_phase",
    "violinplot_algorithm_performance",
    "violinplot_reference_pep",
]

from pepbench.utils._types import str_t


def boxplot_reference_pep(
    data: pd.DataFrame, x: str, y: str | None = "pep_ms", hue: str | None = None, **kwargs: dict
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a boxplot of reference PEP values.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    x : str
        Column name of the x-axis data to plot, typically the different phases or conditions of the experiment.
    y: str, optional
        Column name of the y-axis data to plot. Default: "pep_ms".
    hue: str, optional
        Column name to plot different hues, if desired. Default: None.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot.

    """
    return _plot_helper_reference_pep(data, bp.plotting.feature_boxplot, x, y, hue, **kwargs)


def violinplot_reference_pep(
    data: pd.DataFrame, x: str, y: str | None = "pep_ms", hue: str | None = None, **kwargs: dict
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a violin of reference PEP values.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    x : str
        Column name of the x-axis data to plot, typically the different phases or conditions of the experiment.
    y: str, optional
        Column name of the y-axis data to plot. Default: "pep_ms".
    hue: str, optional
        Column name to plot different hues, if desired. Default: None.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot.

    """
    return _plot_helper_reference_pep(data, sns.violinplot, x, y, hue, **kwargs)


def _plot_helper_reference_pep(
    data: pd.DataFrame,
    plot_func: Callable,
    x: str,
    y: str = "pep_ms",
    hue: str | None = None,
    **kwargs: dict,
) -> tuple[plt.Figure, plt.Axes]:
    fig, ax = _get_fig_ax(kwargs)

    rect = kwargs.pop("rect", (0, 0, 0.85, 1))
    title = kwargs.pop("title", None)

    meanprops = _get_meanprops(**kwargs)

    if hue is None:
        plot_func(data=data.reset_index(), x=x, y=y, ax=ax, meanprops=meanprops, **kwargs)
    else:
        plot_func(data=data.reset_index(), x=x, y=y, hue=hue, meanprops=meanprops, ax=ax, **kwargs)

    ax.set_ylabel(_ylabel_mapping[y])
    ax.set_xlabel(_xlabel_mapping[x])

    if title is not None:
        ax.set_title(title, fontweight="bold")

    fig.tight_layout(rect=rect)
    return fig, ax


def boxplot_algorithm_performance(
    data: pd.DataFrame, metric: str = "absolute_error_per_sample_ms", **kwargs: dict
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a boxplot of a performance metric for different algorithms.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    metric : str, optional
        Algorithm performance metric to plot. Default: "absolute_error_per_sample_ms"
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot.

    """
    return _plot_helper_algorithm_performance(data, sns.boxplot, metric, **kwargs)


def violinplot_algorithm_performance(
    data: pd.DataFrame, metric: str = "absolute_error_per_sample_ms", **kwargs: dict
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a violinplot of a performance metric for different algorithms.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    metric : str, optional
        Algorithm performance metric to plot. Default: "absolute_error_per_sample_ms"
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot.

    """
    return _plot_helper_algorithm_performance(data, sns.violinplot, metric, **kwargs)


def _plot_helper_algorithm_performance(
    data: pd.DataFrame, plot_func: Callable, metric: str = "absolute_error_per_sample_ms", **kwargs: dict
) -> tuple[plt.Figure, plt.Axes]:
    fig, ax = _get_fig_ax(kwargs)
    rect = kwargs.pop("rect", (0, 0, 1, 1))

    data = get_performance_metric(data, metric)

    # create new index level which merges the algorithm levels
    data = data.reset_index()
    data = data.dropna().astype({metric: "float"})
    data = data.replace(_algorithm_mapping)

    algo_levels = [s for s in data.columns if s in _algo_level_mapping]
    data = data.assign(algorithm=data[algo_levels].apply(lambda x: "\n".join(x), axis=1))

    # filter kwargs for sns.boxplot
    if "boxplot" in plot_func.__name__:
        kwargs_plot = {
            k: v
            for k, v in kwargs.items()
            if k
            in list(inspect.signature(sns.boxplot).parameters.keys())
            + list(inspect.signature(plt.boxplot).parameters.keys())
        }
        kwargs_plot["meanprops"] = _get_meanprops(**kwargs)
    elif "violin" in plot_func.__name__:
        kwargs_plot = {
            k: v for k, v in kwargs.items() if k in list(inspect.signature(sns.violinplot).parameters.keys())
        }
    else:
        raise ValueError(f"Unknown plot function: {plot_func.__name__}")

    plot_func(data, x="algorithm", y=metric, hue="algorithm", ax=ax, **kwargs_plot)
    ax.set_ylabel(_metric_mapping[metric])

    xlabel = _format_pep_pipeline(algo_levels)

    ax.set_xlabel(
        xlabel,
        labelpad=12,
        # fontweight="bold",
    )

    if "title" in kwargs:
        ax.set_title(kwargs.pop("title"), fontweight="bold")

    fig.tight_layout(rect=rect)
    return fig, ax


def residual_plot_pep(data: pd.DataFrame, algorithm: str_t, **kwargs: dict) -> tuple[plt.Figure, plt.Axes]:
    """Plot a residual plot of PEP values for a specific algorithm.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    algorithm : str or list of str
        Name of the algorithm (or list of algorithm names, if a pipeline is used) to plot.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("color", cmaps.fau[0])
    kwargs.setdefault("alpha", 0.3)

    show_upper_limit = kwargs.pop("show_upper_limit", False)

    if isinstance(algorithm, str):
        algorithm = [algorithm]

    fig, ax = _get_fig_ax(kwargs)

    algo_levels = [s for s in data.index.names if s in _algo_level_mapping]

    data = get_data_for_algo(data, algorithm)

    data = data["pep_ms"]
    data = data.dropna()

    _plot_blandaltman(x=data["reference"], y=data["estimated"], xaxis="x", ax=ax, **kwargs)

    title = _format_title(algo_levels, algorithm)

    ax.set_xlabel("Reference PEP [ms]")
    ax.set_ylabel("Reference - Estimated PEP [ms]")
    ax.set_title(title, fontdict={"fontweight": "bold"})

    if show_upper_limit:
        xvals = np.arange(0, int(ax.get_ylim()[1]), 1)
        ax.plot(xvals, xvals, color=cmaps.wiso[0], ls="--")

    fig.tight_layout()
    return fig, ax


def residual_plot_pep_participant(data: pd.DataFrame, algorithm: str_t, **kwargs: dict) -> tuple[plt.Figure, plt.Axes]:
    """Plot a residual plot of PEP values for a specific algorithm, grouped by participant.

    Each participant is represented by a different color. The name of the participant column is assumed to be
    "participant".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    algorithm : str or list of str
        Name of the algorithm (or list of algorithm names, if a pipeline is used) to plot.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.


    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("base_color", "Spectral")
    kwargs.setdefault("show_legend", False)
    kwargs.setdefault("rect", (0, 0, 1, 1))
    return _residual_plot_error_detailed_helper(data, algorithm, "participant", **kwargs)


def residual_plot_pep_phase(
    data: pd.DataFrame, algorithm: Sequence[str], **kwargs: dict
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a residual plot of PEP values for a specific algorithm, grouped by experimental phase.

    Each experimental phase is represented by a different color. The name of the phase column is assumed to be "phase".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    algorithm : str or list of str
        Name of the algorithm (or list of algorithm names, if a pipeline is used) to plot.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.


    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("base_color", f"blend:{colors_all.fau},{colors_all.tech_light}")
    return _residual_plot_error_detailed_helper(data, algorithm, "phase", **kwargs)


def residual_plot_pep_heart_rate(
    data: pd.DataFrame, algorithm: Sequence[str], bins: int | str | Sequence[int] | None = 10, **kwargs: dict
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a residual plot of PEP values for a specific algorithm, grouped by heart rate bins.

    Each heart rate bin is represented by a different color. The name of the heart rate column is assumed to be
    "heart_rate_bpm".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    algorithm : str or list of str
        Name of the algorithm (or list of algorithm names, if a pipeline is used) to plot.
    bins : int or str
        Number of bins to use for the heart rate histogram using :func:`~numpy.histogram`. If `bins` is an int,
        it defines the number of equal-width bins in the given range (10, by default). If `bins` is a sequence, it
        defines a monotonically increasing array of bin edges, including the rightmost edge, allowing for
        non-uniform bin widths. If `bins` is a string, it defines the method used to calculate the
        optimal bin width, as defined by :func:`~numpy.histogram_bin_edges`.
        See also :func:`~numpy.histogram` for more information.
        Default: 10.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("rect", (0, 0, 0.85, 1))
    kwargs.setdefault("base_color", "Spectral_r")

    histogram, bin_edges = np.histogram(data["heart_rate_bpm"].dropna(), bins=bins)

    kwargs["num_groups"] = len(bin_edges) - 1

    # add category for heart rate
    data = data.assign(
        heart_rate_range=pd.cut(
            data[("heart_rate_bpm", "estimated")],
            bins=bin_edges,
            labels=[f"{int(bin_edges[i])}-{int(bin_edges[i + 1])}" for i in range(len(bin_edges) - 1)],
        )
    )
    data = data.set_index("heart_rate_range", append=True)
    return _residual_plot_error_detailed_helper(data, algorithm, "heart_rate_range", **kwargs)


def residual_plot_pep_bmi(
    data: pd.DataFrame, algorithm: Sequence[str], bins: int | str | Sequence[int] | None = 4, **kwargs: dict
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a residual plot of PEP values for a specific algorithm, grouped by BMI bins.

    Each BMI bin is represented by a different color. The name of the BMI column is assumed to be "BMI".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    algorithm : str or list of str
        Name of the algorithm (or list of algorithm names, if a pipeline is used) to plot.
    bins : int or str
        Number of bins to use for the heart rate histogram using :func:`~numpy.histogram`. If `bins` is an int,
        it defines the number of equal-width bins in the given range (4, by default). If `bins` is a sequence, it
        defines a monotonically increasing array of bin edges, including the rightmost edge, allowing for
        non-uniform bin widths. If `bins` is a string, it defines the method used to calculate the
        optimal bin width, as defined by :func:`~numpy.histogram_bin_edges`.
        See also :func:`~numpy.histogram` for more information.
        Default: 4.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("rect", (0, 0, 0.85, 1))
    kwargs.setdefault("base_color", "Spectral_r")

    histogram, bin_edges = np.histogram(data["BMI"].dropna(), bins=bins)
    kwargs["num_groups"] = len(bin_edges) - 1

    # add category for heart rate
    data = data.assign(
        bmi_range=pd.cut(
            data[("BMI", "estimated")],
            bins=bin_edges,
            labels=[f"{int(bin_edges[i])}-{int(bin_edges[i + 1])}" for i in range(len(bin_edges) - 1)],
        )
    )
    data = data.set_index("bmi_range", append=True)
    return _residual_plot_error_detailed_helper(data, algorithm, "bmi_range", **kwargs)


def residual_plot_pep_age(
    data: pd.DataFrame, algorithm: Sequence[str], bins: int | str | Sequence[int] | None = 5, **kwargs: dict
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a residual plot of PEP values for a specific algorithm, grouped by age bins.

    Each age bin is represented by a different color. The name of the age column is assumed to be "Age".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    algorithm : str or list of str
        Name of the algorithm (or list of algorithm names, if a pipeline is used) to plot.
    bins : int or str
        Number of bins to use for the age histogram using :func:`~numpy.histogram`. If `bins` is an int,
        it defines the number of equal-width bins in the given range (5, by default). If `bins` is a sequence, it
        defines a monotonically increasing array of bin edges, including the rightmost edge, allowing for
        non-uniform bin widths. If `bins` is a string, it defines the method used to calculate the
        optimal bin width, as defined by :func:`~numpy.histogram_bin_edges`.
        See also :func:`~numpy.histogram` for more information.
        Default: 5.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("rect", (0, 0, 0.85, 1))
    kwargs.setdefault("base_color", "Spectral_r")

    histogram, bin_edges = np.histogram(data["Age"].dropna(), bins=bins)
    kwargs["num_groups"] = len(bin_edges) - 1

    # add category for age
    data = data.assign(
        age_range=pd.cut(
            data[("Age", "estimated")],
            bins=bin_edges,
            labels=[f"{int(bin_edges[i])}-{int(bin_edges[i + 1])}" for i in range(len(bin_edges) - 1)],
        )
    )
    data = data.set_index("age_range", append=True)
    return _residual_plot_error_detailed_helper(data, algorithm, "age_range", **kwargs)


def _residual_plot_error_detailed_helper(
    data: pd.DataFrame, algorithm: str_t, grouper: str, **kwargs: dict
) -> tuple[plt.Figure, plt.Axes]:
    kwargs.setdefault("alpha", 0.8)
    rect = kwargs.pop("rect", (0, 0, 0.90, 1))
    # create a new color palette based on the base color with the length of the number of groups
    n_colors = kwargs.pop("num_groups", data.index.get_level_values(grouper).nunique())
    base_color = kwargs.pop("base_color", "Spectral")
    palette = sns.color_palette(base_color, n_colors=n_colors)
    show_upper_limit = kwargs.pop("show_upper_limit", False)
    show_legend = kwargs.pop("show_legend", True)

    fig, ax = _get_fig_ax(kwargs)

    # use residual plot to only plot mean and confidence interval of all data;
    # manually plot the scatter plot using participant as hue variable afterwards
    kwargs_new = kwargs.copy()
    kwargs_new.update(alpha=0.0)

    data_scatter = get_data_for_algo(data, algorithm)

    data_scatter = data_scatter["pep_ms"].reset_index()
    data_scatter = data_scatter.assign(
        x=data_scatter["reference"], y=data_scatter["reference"] - data_scatter["estimated"]
    )

    # filter kwargs for plt.scatter
    kwargs_scatter = {
        k: v
        for k, v in kwargs.items()
        if k
        in list(inspect.signature(sns.scatterplot).parameters.keys())
        + list(inspect.signature(plt.scatter).parameters.keys())
    }

    sns.scatterplot(data=data_scatter, x="x", y="y", hue=grouper, ax=ax, **kwargs_scatter, palette=palette)

    fig, ax = residual_plot_pep(data, algorithm, ax=ax, **kwargs_new)

    if show_upper_limit:
        xvals = np.arange(0, int(ax.get_ylim()[1]), 1)
        ax.plot(xvals, xvals, color=cmaps.wiso[0], ls="--")

    handles, labels = ax.get_legend_handles_labels()
    ax.legend().remove()
    handles, labels = _remove_duplicate_legend_entries(handles, labels)
    if show_legend:
        fig.legend(
            handles=handles,
            labels=labels,
            title=" ".join([s.capitalize() for s in grouper.split("_")]),
            loc="upper right",
        )

    fig.tight_layout(rect=rect)

    return fig, ax


def _add_corr_coeff(data: pd.DataFrame, x: str, y: str, ax: plt.Axes, **kwargs: dict) -> None:
    kwargs.setdefault("x_coord", 0.95)
    kwargs.setdefault("y_coord", 0.95)

    corr = pg.corr(data[x], data[y])
    s = f"r = {corr['r'].iloc[0]:.2f}, p {_format_p_value(corr['p-val'].iloc[0])}"
    prefix = kwargs.get("prefix")
    if prefix:
        s = f"{prefix}: " + s

    ax.text(
        x=kwargs["x_coord"],
        y=kwargs["y_coord"],
        s=s,
        transform=ax.transAxes,
        fontsize="medium",
        verticalalignment="top",
        horizontalalignment="right",
    )


def histplot_heart_rate(data: pd.DataFrame, hue: str | None = None, **kwargs: dict) -> tuple[plt.Figure, plt.Axes]:
    """Plot a histogram of heart rate values.

    The heart rate is assumed to be in the column "heart_rate_bpm".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    hue : str, optional
        Column name to plot different hues, if desired. Default: None.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function. See :func:`~seaborn.histplot` for more
        information.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("stat", "percent")
    kwargs.setdefault("kde", True)
    show_legend = kwargs.pop("show_legend", True)
    # legend_loc = kwargs.pop("legend_loc", "upper right")

    rect_default = (0, 0, 0.85, 1) if show_legend else (0, 0, 1, 1)

    rect = kwargs.pop("rect", rect_default)

    fig, ax = _get_fig_ax(kwargs)
    ax = sns.histplot(data=data.reset_index(), x="heart_rate_bpm", hue=hue, ax=ax, **kwargs)

    ax.set_xlabel("Heart Rate [bpm]")

    fig.tight_layout(rect=rect)
    # if show_legend and hue is not None:
    #    fig.legend(title=hue.capitalize(), handles=handles, labels=labels, loc=legend_loc)

    return fig, ax


def regplot_pep_heart_rate(
    data: pd.DataFrame,
    algorithm: str_t | None = None,
    use_reference: bool = False,
    add_corr_coeff: bool = False,
    groupby: str | None = None,
    **kwargs: dict,
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a regression plot of PEP values against heart rate.

    The PEP values are assumed to be in the column "pep_ms" and the heart rate values are assumed to be in the column
    "heart_rate_bpm".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    algorithm : str or list of str, optional
        Name of the algorithm (or list of algorithm names, if a pipeline is used) to plot.
        If ``use_reference`` is True, this parameter is ignored. If ``use_reference`` is False,
        this parameter is required.
    use_reference : bool, optional
        ``True`` to use the reference data, ``False`` to use the results of the algorithm (pipeline)
        specified in ``algorithm``. Default: ``False``.
    add_corr_coeff : bool, optional
        ``True`` to add the correlation coefficient to the plot, ``False`` otherwise. Default: ``False``.
    groupby : str, optional
        Column name to group the data by. If specified, the data will be grouped by this column and a repeated-measures
        regression plot will be created using :func:`~pingouin.plot_rm_corr`. ``None`` by default, which means that no
        grouping is applied and a standard regression plot is created using :func:`~seaborn.regplot`.

    kwargs : Any
        Additional keyword arguments to pass to the plotting function. See :func:`~seaborn.regplot` for more
        information.

    Raises
    ------
    ValueError
        If ``use_reference`` is False and ``algorithm`` is not specified.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    if groupby is None:
        kwargs.setdefault("color", cmaps.tech[0])
        kwargs.setdefault("line_kws", {"color": cmaps.tech_dark[0], "alpha": 0.8})

    kwargs.setdefault("scatter_kws", {"alpha": 0.4})
    fig, ax = _get_fig_ax(kwargs)

    if use_reference:
        data = get_reference_data(data)
    else:
        if algorithm is None:
            raise ValueError("If `use_reference` is False, `algorithm` must be specified.")
        data = get_data_for_algo(data, algorithm)

    if data.columns.nlevels > 1:
        data = data.xs("estimated", level=-1, axis=1)

    # filter kwargs for sns.regplot
    kwargs_regplot = {
        k: v
        for k, v in kwargs.items()
        if k
        in list(inspect.signature(sns.regplot).parameters.keys())
        + list(inspect.signature(plt.scatter).parameters.keys())
        + list(inspect.signature(plt.plot).parameters.keys())
    }

    if groupby is not None:
        palette = iter(kwargs.get("palette", cmaps.faculties_light))
        palette_line = iter(kwargs.get("palette_line", cmaps.faculties_dark))
        data.groupby(groupby, group_keys=False).apply(
            lambda df: sns.regplot(
                data=df.reset_index(),
                x="heart_rate_bpm",
                y="pep_ms",
                ax=ax,
                color=next(palette),
                line_kws={"color": next(palette_line), "alpha": 0.8},
                **kwargs_regplot,
                label=df.name,
            )
        )

        if add_corr_coeff:
            handles, labels = ax.get_legend_handles_labels()
            corr_vals = {}
            for name, group in data.groupby(groupby):
                corr = pg.corr(group["heart_rate_bpm"], group["pep_ms"])
                corr_vals[name] = f"r = {corr['r'].iloc[0]:.2f}, p {_format_p_value(corr['p-val'].iloc[0])}"

            labels = [f"{name}: {corr_vals[name]}" for name in labels]
            fig.legend(
                handles=handles,
                labels=labels,
                title=kwargs.get("legend_title", groupby.capitalize()),
                loc=kwargs.get("legend_loc", "upper right"),
            )
    else:
        sns.regplot(data=data.reset_index(), x="heart_rate_bpm", y="pep_ms", ax=ax, **kwargs_regplot)
        if add_corr_coeff:
            _add_corr_coeff(data, x="heart_rate_bpm", y="pep_ms", ax=ax)

    ax.set_xlabel("Heart Rate [bpm]")
    ax.set_ylabel("PEP [ms]")

    title = "PEP Reference" if use_reference else "PEP Pipeline:\n" + _pep_pipeline_to_str(algorithm)
    ax.set_title(title, fontweight="bold")

    fig.tight_layout(rect=kwargs.get("rect", (0, 0, 1, 1)))

    return fig, ax


def regplot_error_heart_rate(
    data: pd.DataFrame,
    algorithm: str_t,
    error_metric: str = "error_per_sample_ms",
    add_corr_coeff: bool = False,
    **kwargs: dict,
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a regression plot of a PEP estimation *error* metric against heart rate.

    The error metric is assumed to be in the column specified by the parameter ``error_metric`` and the heart rate
    values are assumed to be in the column "heart_rate_bpm".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    algorithm : str or list of str
        Name of the algorithm (or list of algorithm names, if a pipeline is used) to plot.
    error_metric : str, optional
        Error metric to plot. Default: "error_per_sample_ms".
    add_corr_coeff : bool, optional
        ``True`` to add the correlation coefficient to the plot, ``False`` otherwise. Default: ``False``.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function. See :func:`~seaborn.regplot` for more
        information.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("color", cmaps.tech[0])
    kwargs.setdefault("scatter_kws", {"alpha": 0.3})
    kwargs.setdefault("line_kws", {"color": cmaps.fau[0], "alpha": 0.8})
    fig, ax = _get_fig_ax(kwargs)

    if isinstance(algorithm, str):
        algorithm = [algorithm]
    algo_levels = [s for s in data.index.names if s in _algo_level_mapping]

    data = get_data_for_algo(data, algorithm)

    data = data.reindex(["estimated", "metric"], level=-1, axis=1)
    data = data.droplevel(level=-1, axis=1)

    sns.regplot(data=data.reset_index(), x="heart_rate_bpm", y=error_metric, ax=ax, **kwargs)

    if add_corr_coeff:
        _add_corr_coeff(data, x="heart_rate_bpm", y=error_metric, ax=ax)

    ax.set_xlabel("Heart Rate [bpm]")
    ax.set_ylabel(_ylabel_mapping[error_metric])

    if len(algo_levels) == 1:
        title = f"{_algo_level_mapping[algo_levels[0]]}: {_pep_pipeline_to_str(algorithm)}"
    else:
        title = f"PEP Pipeline:\n{_pep_pipeline_to_str(algorithm)}"
    ax.set_title(title, fontweight="bold")

    return fig, ax


def regplot_error_bmi(
    data: pd.DataFrame,
    algorithm: str_t,
    error_metric: str = "error_per_sample_ms",
    add_corr_coeff: bool = False,
    **kwargs: dict,
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a regression plot of a PEP estimation *error* metric against BMI.

    The error metric is assumed to be in the column specified by the parameter ``error_metric`` and the BMI values
    are assumed to be in the column "BMI".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    algorithm : str or list of str
        Name of the algorithm (or list of algorithm names, if a pipeline is used) to plot.
    error_metric : str, optional
        Error metric to plot. Default: "error_per_sample_ms".
    add_corr_coeff : bool, optional
        ``True`` to add the correlation coefficient to the plot, ``False`` otherwise. Default: ``False``.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function. See :func:`~seaborn.regplot` for more
        information.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("color", cmaps.tech[0])
    kwargs.setdefault("scatter_kws", {"alpha": 0.3})
    kwargs.setdefault("line_kws", {"color": cmaps.fau[0], "alpha": 0.8})
    fig, ax = _get_fig_ax(kwargs)

    if isinstance(algorithm, str):
        algorithm = [algorithm]
    algo_levels = [s for s in data.index.names if s in _algo_level_mapping]

    data = get_data_for_algo(data, algorithm)

    data = data.reindex(["estimated", "metric"], level=-1, axis=1)
    data = data.droplevel(level=-1, axis=1)

    sns.regplot(data=data.reset_index(), x="BMI", y=error_metric, ax=ax, **kwargs)

    if add_corr_coeff:
        _add_corr_coeff(data, x="BMI", y=error_metric, ax=ax)

    ax.set_xlabel("BMI [kg/m²]")
    ax.set_ylabel(_ylabel_mapping[error_metric])

    if len(algo_levels) == 1:
        title = f"{_algo_level_mapping[algo_levels[0]]}: {_pep_pipeline_to_str(algorithm)}"
    else:
        title = f"PEP Pipeline:\n{_pep_pipeline_to_str(algorithm)}"
    ax.set_title(title, fontweight="bold")

    return fig, ax


def regplot_error_age(
    data: pd.DataFrame,
    algorithm: str_t,
    error_metric: str = "error_per_sample_ms",
    add_corr_coeff: bool = False,
    **kwargs: dict,
) -> tuple[plt.Figure, plt.Axes]:
    """Plot a regression plot of a PEP estimation *error* metric against age.

    The error metric is assumed to be in the column specified by the parameter ``error_metric`` and the age values
    are assumed to be in the column "age".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    algorithm : str or list of str
        Name of the algorithm (or list of algorithm names, if a pipeline is used) to plot.
    error_metric : str, optional
        Error metric to plot. Default: "error_per_sample_ms".
    add_corr_coeff : bool, optional
        ``True`` to add the correlation coefficient to the plot, ``False`` otherwise. Default: ``False``.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function. See :func:`~seaborn.regplot` for more
        information.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("color", cmaps.tech[0])
    kwargs.setdefault("scatter_kws", {"alpha": 0.3})
    kwargs.setdefault("line_kws", {"color": cmaps.fau[0], "alpha": 0.8})
    fig, ax = _get_fig_ax(kwargs)

    if isinstance(algorithm, str):
        algorithm = [algorithm]
    algo_levels = [s for s in data.index.names if s in _algo_level_mapping]

    data = get_data_for_algo(data, algorithm)

    data = data.reindex(["estimated", "metric"], level=-1, axis=1)
    data = data.droplevel(level=-1, axis=1)

    sns.regplot(data=data.reset_index(), x="Age", y=error_metric, ax=ax, **kwargs)

    if add_corr_coeff:
        _add_corr_coeff(data, x="Age", y=error_metric, ax=ax)

    ax.set_xlabel("Age [years]")
    ax.set_ylabel(_ylabel_mapping[error_metric])

    if len(algo_levels) == 1:
        title = f"{_algo_level_mapping[algo_levels[0]]}: {_pep_pipeline_to_str(algorithm)}"
    else:
        title = f"PEP Pipeline:\n{_pep_pipeline_to_str(algorithm)}"
    ax.set_title(title, fontweight="bold")

    return fig, ax


def paired_plot_error_outlier_correction(
    data: pd.DataFrame, outlier_algo_combis: Sequence[Sequence[str]], dv: str, **kwargs: dict[str, Any]
) -> tuple[plt.Figure, Sequence[plt.Axes]]:
    """Plot a paired plot of an error metric for different outlier correction algorithms.

    A paired plot is a repeated-measures boxplot that additionally shows the paired data points and highlights the
    changes between the paired data points. This plot is useful to visualize the effect of different outlier correction
    algorithms on the error metric.

    See :func:``pingouin.plot_paired`` for more information.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    outlier_algo_combis : list of string tuples
        List of tuples containing algorithm pipeline steps to compare against
    dv : str
        Dependent variable to plot. Can either be an error metric or a performance metric.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, list of :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("ncols", len(outlier_algo_combis))
    fig, axs = _get_fig_axs(kwargs)

    if "error" in dv:
        colors = kwargs.pop("colors", ["indianred", "grey", "green"])
    else:
        colors = kwargs.pop("colors", ["grey", "green", "indianred"])
    pointplot_kwargs = kwargs.pop("pointplot_kwargs", {"scale": 0.6, "marker": ".", "alpha": 0.5})

    for i, (ax, outlier_combi) in enumerate(zip(axs, outlier_algo_combis, strict=False)):
        data_plot = data.reindex(outlier_combi, level="outlier_correction_algorithm")
        data_plot = data_plot.unstack("outlier_correction_algorithm")
        eq_mask = ~(data_plot.diff(axis=1) == 0).any(axis=1)
        data_plot = data_plot.loc[eq_mask].stack(future_stack=True)

        # rename outlier correction algorithms
        data_plot = data_plot.rename(index=_algorithm_mapping)
        outlier_combi_format = [_algorithm_mapping[algo] for algo in outlier_combi]
        _plot_paired(
            data=data_plot,
            dv=dv,
            within="outlier_correction_algorithm",
            subject="id_concat",
            order=outlier_combi_format,
            colors=colors,
            pointplot_kwargs=pointplot_kwargs,
            ax=ax,
        )

        if i == 0:
            ax.set_ylabel(_ylabel_mapping[dv])
        ax.set_xlabel(_algo_level_mapping["outlier_correction_algorithm"])

        # xmax = len(outlier_combi) - 1 + 0.15
        ax.set_xlim([-0.15, 1.15])
    if "title" in kwargs:
        fig.suptitle(f"B-Point Algorithm: {_algorithm_mapping[kwargs.pop('title')]}", fontweight="bold")

    fig.tight_layout()

    return fig, axs


def paired_plot_error_pep_pipeline(
    data: pd.DataFrame, pep_pipelines: Sequence[Sequence[str]], dv: str, **kwargs: dict[str, Any]
) -> tuple[plt.Figure, Sequence[plt.Axes]]:
    """Plot a paired plot of an error metric for different PEP extraction pipelines.

    A paired plot is a repeated-measures boxplot that additionally shows the paired data points and highlights the
    changes between the paired data points. This plot is useful to visualize the differences between different PEP
    extraction pipelines.

    See :func:``pingouin.plot_paired`` for more information.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Dataframe containing the data to plot.
    pep_pipelines : list of string tuples
        List of tuples with pipeline configurations to compare against
    dv : str
        Dependent variable to plot. Can either be an error metric or a performance metric.
    kwargs : Any
        Additional keyword arguments to pass to the plotting function.

    Returns
    -------
    :class:`~matplotlib.figure.Figure`, list of :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    kwargs.setdefault("ncols", len(pep_pipelines))
    fig, axs = _get_fig_axs(kwargs)
    if "error" in dv:
        colors = kwargs.pop("colors", ["#C50F3C", "#8C9FB1", "#7BB725"])
    else:
        colors = kwargs.pop("colors", ["#8C9FB1", "#7BB725", "#C50F3C"])
    pointplot_kwargs = kwargs.pop("pointplot_kwargs", {"scale": 0.6, "marker": ".", "alpha": 0.2})

    for i, (ax, pep_pipeline) in enumerate(zip(axs, pep_pipelines, strict=False)):
        pep_pipeline_format = ["_".join(p) for p in pep_pipeline]
        data_plot = data.reindex(pep_pipeline_format, level="pipeline")
        data_plot = data_plot.unstack("pipeline")
        eq_mask = ~(data_plot.diff(axis=1) == 0).any(axis=1)
        data_plot = data_plot.loc[eq_mask].stack(future_stack=True)

        # rename outlier correction algorithms
        _plot_paired(
            data=data_plot,
            dv=dv,
            within="pipeline",
            subject="id_concat",
            order=pep_pipeline_format,
            colors=colors,
            pointplot_kwargs=pointplot_kwargs,
            ax=ax,
        )

        if i == 0:
            ax.set_ylabel(_ylabel_mapping[dv])
        # ax.set_xlabel(_algo_level_mapping["outlier_correction_algorithm"])
        xmax = len(pep_pipeline_format) - 1 + 0.15
        ax.set_xlim([-0.15, xmax])
    if "title" in kwargs:
        fig.suptitle(f"B-Point Algorithm: {_algorithm_mapping[kwargs.pop('title')]}", fontweight="bold")

    fig.tight_layout()

    return fig, axs


def _get_meanprops(**kwargs: dict) -> dict:
    if "meanprops" in kwargs:
        return kwargs["meanprops"]
    return {"marker": "X", "markerfacecolor": "white", "markeredgecolor": "black", "markersize": "6"}


def _format_title(algo_levels: Sequence[str], algorithm: Sequence[str]) -> str:
    if len(algo_levels) == 1:
        title = f"{_algo_level_mapping[algo_levels[0]]}: {_pep_pipeline_to_str(algorithm)}"
    else:
        title = f"PEP Pipeline:\n{_pep_pipeline_to_str(algorithm)}"
    return title


def _pep_pipeline_to_str(pipeline: Sequence[str]) -> str:
    return " | ".join([_algorithm_mapping[algo] for algo in pipeline])


def _format_p_value(p_val: float) -> str:
    # sanitize p-value to ensure it's a float, then format it to 3 decimal places
    p_val = f"{float(p_val):.3f}"
    p_val = "< 0.001" if p_val == "0.000" else "= " + p_val
    return p_val


def _format_pep_pipeline(algo_levels: Sequence[str]) -> str:
    if algo_levels == 3:
        pipeline_str = f"PEP Pipeline ({' | '.join([_algo_level_mapping[algo_level] for algo_level in algo_levels])})"
    elif algo_levels == 1:
        pipeline_str = _algo_level_mapping[algo_levels[0]]
    else:
        pipeline_str = " | ".join([_algo_level_mapping[algo_level] for algo_level in algo_levels])

    return pipeline_str


def plot_q_wave_detection_waveform_detailed_comparison(
    datapoint_01: pd.DataFrame,
    datapoint_02: pd.DataFrame,
    base_plot_func: Callable,
    plot_func_01_params: dict,
    plot_func_02_params: dict,
    ax_inset_01_params: dict,
    ax_inset_02_params: dict,
    datapoint_01_name: str,
    datapoint_02_name: str,
) -> plt.Figure | Sequence[plt.Axes]:
    """Plot a detailed comparison of two Q-wave detection waveforms with insets.

    This function creates a figure with two subplots, each containing a waveform plot of a Q-wave detection result.
    Additionally, each subplot contains an inset with a zoomed-in view of the waveform. This function can be helpful
    to compare the effect of different ECG waveforms on the Q-wave detection result.

    Parameters
    ----------
    datapoint_01 : :class:`~pandas.DataFrame`
        Dataframe containing the first data point to plot.
    datapoint_02 : :class:`~pandas.DataFrame`
        Dataframe containing the second data point to plot.
    base_plot_func : Callable
        Base plotting function to use for the waveform plot. Should be one of the Q-wave plotting functions from
        ``pepbench.plotting.algorithms``.
    plot_func_01_params : dict
        Additional keyword arguments to pass to the plotting function for the first datapoint. Can include parameters
        such as ``normalize_time`` or ``use_tight``.
    plot_func_02_params : dict
        Additional keyword arguments to pass to the plotting function for the second datapoint. Can include parameters
        such as ``normalize_time`` or ``use_tight``.
    ax_inset_01_params : dict
        Parameters to pass for configuring the inset for the first datapoint.
        See :func:`~matplotlib.axes.Axes.inset_axes` for more information.
    ax_inset_02_params : dict
        Parameters to pass for configuring the inset for the second datapoint.
        See :func:`~matplotlib.axes.Axes.inset_axes` for more information.
    datapoint_01_name : str
        Name of the first datapoint to display in the plot.
    datapoint_02_name : str
        Name of the second datapoint to display in the plot.


    Returns
    -------
    :class:`~matplotlib.figure.Figure`, list of :class:`~matplotlib.axes.Axes`
        Figure and axes of the plot

    """
    fig, axs = plt.subplots(
        nrows=2,
        gridspec_kw={"left": 0.075, "bottom": 0.1, "top": 0.85, "right": 0.75, "hspace": 0.25},
        sharex=True,
        sharey=True,
    )

    plot_func_01_params.setdefault("normalize_time", True)
    plot_func_02_params.setdefault("normalize_time", True)
    plot_func_01_params.setdefault("use_tight", False)
    plot_func_02_params.setdefault("use_tight", False)

    base_plot_func(datapoint=datapoint_01, ax=axs[0], **plot_func_01_params)
    base_plot_func(datapoint=datapoint_02, ax=axs[1], **plot_func_02_params)

    xlim = axs[0].get_xlim()
    ylim = axs[0].get_ylim()

    xmin_01 = ax_inset_01_params.pop("xmin", 0)
    xmax_01 = ax_inset_01_params.pop("xmax", 1)

    ax_inset1 = axs[0].inset_axes(
        **ax_inset_01_params, yticks=[], yticklabels=[], xlim=(xmin_01, xmax_01), ylim=(ylim[0] + 0.1, ylim[1] - 0.1)
    )
    base_plot_func(datapoint=datapoint_01, ax=ax_inset1, **plot_func_01_params)

    ax_inset1.tick_params(axis="both", length=0, labelbottom=False, labelleft=False)
    ax_inset1.set_xlim(xmin_01, xmax_01)
    ax_inset1.set_xlabel(None)
    ax_inset1.set_ylabel(None)

    xmin_02 = ax_inset_02_params.pop("xmin", 0)
    xmax_02 = ax_inset_02_params.pop("xmax", 1)

    ax_inset2 = axs[1].inset_axes(
        **ax_inset_02_params, yticks=[], yticklabels=[], xlim=(xmin_02, xmax_02), ylim=(ylim[0] + 0.1, ylim[1] - 0.1)
    )
    base_plot_func(datapoint=datapoint_02, ax=ax_inset2, **plot_func_02_params)

    ax_inset2.tick_params(axis="both", length=0, labelbottom=False, labelleft=False)
    ax_inset2.set_xlim(xmin_02, xmax_02)
    ax_inset2.set_xlabel(None)
    ax_inset2.set_ylabel(None)

    axs[0].set_xlim(0.1, xlim[0] + 0.33 * (xlim[1] - xlim[0]))

    axs[0].indicate_inset_zoom(ax_inset1, edgecolor="black")
    axs[1].indicate_inset_zoom(ax_inset2, edgecolor="black")

    axs[0].text(x=0.5, y=1.05, s=datapoint_01_name, transform=axs[0].transAxes, fontdict={"fontweight": "bold"})
    axs[1].text(x=0.5, y=1.05, s=datapoint_02_name, transform=axs[1].transAxes, fontdict={"fontweight": "bold"})

    return fig, axs



"""Module for ECG event extraction algorithms."""

from biopsykit.signals.ecg.event_extraction import (
    QPeakExtractionForouzanfar2018,
    QPeakExtractionMartinez2004Neurokit,
    QPeakExtractionVanLien2013,
)

__all__ = [
    "QPeakExtractionForouzanfar2018",
    "QPeakExtractionMartinez2004Neurokit",
    "QPeakExtractionVanLien2013",
]



from collections.abc import Sequence
from typing import Any

import numpy as np
import pandas as pd
import seaborn as sns
from biopsykit.signals._base_extraction import BaseExtraction
from biopsykit.signals.ecg.event_extraction import BaseEcgExtractionWithHeartbeats
from biopsykit.signals.icg.event_extraction import BaseBPointExtraction, CPointExtractionScipyFindPeaks
from fau_colors import cmaps
from matplotlib import pyplot as plt
from matplotlib import transforms
from scipy import stats

from pepbench.datasets import BasePepDataset, BasePepDatasetWithAnnotations
from pepbench.plotting._utils import (
    _add_ecg_q_peak_artefacts,
    _add_ecg_q_peaks,
    _add_heartbeat_borders,
    _add_icg_b_point_artefacts,
    _add_icg_b_points,
    _add_pep_from_reference,
    _add_pep_from_results,
    _get_bbox_coords,
    _get_data,
    _get_fig_ax,
    _get_fig_axs,
    _get_heartbeats,
    _get_labels_from_challenge_results,
    _get_legend_loc,
    _get_rect,
    _get_reference_labels,
    _handle_legend_one_axis,
    _handle_legend_two_axes,
    _sanitize_heartbeat_subset,
    add_fancy_patch_around,
)

__all__ = [
    "_plot_blandaltman",
    "_plot_paired",
    "plot_signals",
    "plot_signals_from_challenge_results",
    "plot_signals_with_algorithm_results",
    "plot_signals_with_reference_labels",
    "plot_signals_with_reference_pep",
]


def plot_signals(
    datapoint: BasePepDataset,
    *,
    collapse: bool = False,
    normalize_time: bool = False,
    heartbeat_subset: Sequence[int] | None = None,
    **kwargs: Any,
) -> tuple[plt.Figure, plt.Axes | Sequence[plt.Axes]]:
    """Plot ECG and ICG signals.

    Parameters
    ----------
    datapoint : :class:`~pepbench.datasets.BasePepDataset`
        Dataset to plot.
    collapse : bool, optional
        If ``True``, plot ECG and ICG signals in one axis. If ``False``, plot ECG and ICG signals in two axes.
        Default: ``False``.
    normalize_time : bool, optional
        If ``True``, normalize time to seconds. If ``False``, use the original time format.
        Default: ``False``.
    heartbeat_subset : list of int, optional
        List of heartbeats (as indices) to plot. If ``None``, plot all heartbeats.
        Default: ``None``.
    kwargs : dict, optional
        Additional keyword arguments.

    """
    if collapse:
        return _plot_signals_one_axis(
            datapoint=datapoint,
            normalize_time=normalize_time,
            heartbeat_subset=heartbeat_subset,
            **kwargs,
        )
    return _plot_signals_two_axes(
        datapoint=datapoint,
        normalize_time=normalize_time,
        heartbeat_subset=heartbeat_subset,
        **kwargs,
    )


def plot_signals_with_reference_labels(  # noqa: C901
    datapoint: BasePepDatasetWithAnnotations,
    *,
    heartbeat_subset: Sequence[int] | None = None,
    collapse: bool = False,
    normalize_time: bool = False,
    **kwargs: Any,
) -> tuple[plt.Figure, plt.Axes | Sequence[plt.Axes]]:
    kwargs.setdefault("sharex", True)
    kwargs.setdefault("legend_max_cols", 6)
    kwargs.setdefault("legend_loc", _get_legend_loc(kwargs))
    plot_ecg = kwargs.get("plot_ecg", True)
    plot_icg = kwargs.get("plot_icg", True)
    plot_artefacts = kwargs.get("plot_artefacts", False)
    rect = _get_rect(kwargs)

    fig, ax = plot_signals(
        datapoint,
        collapse=collapse,
        normalize_time=normalize_time,
        heartbeat_subset=heartbeat_subset,
        **kwargs,
    )
    ecg_data, icg_data = _get_data(datapoint, normalize_time=normalize_time, heartbeat_subset=heartbeat_subset)

    reference_labels = _get_reference_labels(datapoint, heartbeat_subset=heartbeat_subset)
    heartbeats = reference_labels["heartbeats"]
    q_peaks = reference_labels["q_peaks"]
    q_peak_artefacts = reference_labels["q_peak_artefacts"]
    b_points = reference_labels["b_points"]
    b_point_artefacts = reference_labels["b_point_artefacts"]

    # plot q-peak onsets and b-points
    if collapse:
        _add_heartbeat_borders(ecg_data.index[list(heartbeats["start_sample"])], ax, **kwargs)
        if plot_ecg:
            _add_ecg_q_peaks(ecg_data, q_peaks, ax, **kwargs)
        if plot_icg:
            _add_icg_b_points(icg_data, b_points, ax, **kwargs)
        if plot_artefacts:
            if not q_peak_artefacts.empty:
                _add_ecg_q_peak_artefacts(ecg_data, q_peak_artefacts, ax, **kwargs)
            if not b_point_artefacts.empty:
                _add_icg_b_point_artefacts(icg_data, b_point_artefacts, ax, **kwargs)

        _handle_legend_one_axis(fig, ax, **kwargs)
    else:
        _add_heartbeat_borders(ecg_data.index[list(heartbeats["start_sample"])], ax[0], **kwargs)
        _add_heartbeat_borders(ecg_data.index[list(heartbeats["start_sample"])], ax[1], **kwargs)
        if plot_ecg:
            _add_ecg_q_peaks(ecg_data, q_peaks, ax[0], **kwargs)
        if plot_icg:
            _add_icg_b_points(icg_data, b_points, ax[1], **kwargs)
        if plot_artefacts:
            if not q_peak_artefacts.empty:
                _add_ecg_q_peak_artefacts(ecg_data, q_peak_artefacts, ax[0], **kwargs)
            if not b_point_artefacts.empty:
                _add_icg_b_point_artefacts(icg_data, b_point_artefacts, ax[1], **kwargs)

        _handle_legend_two_axes(fig, ax, **kwargs)

    fig.tight_layout(rect=rect)
    return fig, ax


def plot_signals_with_reference_pep(
    datapoint: BasePepDatasetWithAnnotations,
    *,
    collapse: bool = False,
    normalize_time: bool = False,
    heartbeat_subset: Sequence[int] | None = None,
    **kwargs: Any,
) -> tuple[plt.Figure, plt.Axes]:
    kwargs.setdefault("legend_orientation", "vertical")
    kwargs.setdefault("legend_outside", False)
    kwargs.setdefault("legend_max_cols", 6)
    kwargs.setdefault("legend_loc", _get_legend_loc(kwargs))
    rect = _get_rect(kwargs)

    fig, axs = plot_signals_with_reference_labels(
        datapoint,
        collapse=collapse,
        normalize_time=normalize_time,
        heartbeat_subset=heartbeat_subset,
        **kwargs,
    )

    ecg_data, icg_data = _get_data(datapoint, normalize_time=normalize_time, heartbeat_subset=heartbeat_subset)

    reference_labels = _get_reference_labels(datapoint, heartbeat_subset=heartbeat_subset)

    reference_labels_ecg = (
        pd.concat(
            [reference_labels["q_peaks"], reference_labels["q_peak_artefacts"]], names=["sample_relative", "label"]
        )
        .sort_index()
        .reset_index()
    )
    reference_labels_icg = (
        pd.concat(
            [reference_labels["b_points"], reference_labels["b_point_artefacts"]], names=["sample_relative", "label"]
        )
        .sort_index()
        .reset_index()
    )
    reference_labels_combined = pd.concat({"ecg": reference_labels_ecg, "icg": reference_labels_icg}, axis=1)

    if collapse:
        _add_pep_from_reference(ecg_data, icg_data, reference_labels_combined, axs, **kwargs)
        _handle_legend_one_axis(fig, axs, **kwargs)
    else:
        _add_pep_from_reference(ecg_data, icg_data, reference_labels_combined, axs[0], **kwargs)
        _add_pep_from_reference(ecg_data, icg_data, reference_labels_combined, axs[1], **kwargs)
        _handle_legend_two_axes(fig, axs, **kwargs)

    fig.tight_layout(rect=rect)
    return fig, axs


def plot_signals_with_algorithm_results(
    datapoint: BasePepDatasetWithAnnotations,
    *,
    collapse: bool = False,
    algorithm: BaseExtraction,
    normalize_time: bool = False,
    heartbeat_subset: Sequence[int] | None = None,
    **kwargs: Any,
) -> tuple[plt.Figure, plt.Axes | Sequence[plt.Axes]]:
    kwargs.setdefault("legend_loc", _get_legend_loc(kwargs))
    kwargs.setdefault("legend_max_cols", 5)
    rect = _get_rect(kwargs)

    fig, axs = plot_signals_with_reference_labels(
        datapoint,
        collapse=collapse,
        normalize_time=normalize_time,
        heartbeat_subset=heartbeat_subset,
        b_point_label="Reference B-Points",
        **kwargs,
    )

    heartbeat_subset = _sanitize_heartbeat_subset(heartbeat_subset)
    ecg_data, icg_data = _get_data(datapoint, normalize_time=normalize_time, heartbeat_subset=heartbeat_subset)
    heartbeats = _get_heartbeats(datapoint, heartbeat_subset)

    if isinstance(algorithm, BaseEcgExtractionWithHeartbeats):
        algorithm.extract(
            ecg=ecg_data,
            heartbeats=heartbeats,
            sampling_rate_hz=datapoint.sampling_rate_ecg,
        )
        q_peaks = algorithm.points_["q_peak_sample"]
        q_peaks = q_peaks.loc[heartbeats.index].dropna()

        if collapse:
            _add_ecg_q_peaks(
                ecg_data,
                q_peaks,
                axs,
                q_peak_label="Detected Q-Peaks",
                q_peak_color=cmaps.med_dark[0],
                **kwargs,
            )
        else:
            _add_ecg_q_peaks(
                ecg_data,
                q_peaks,
                axs[0],
                q_peak_label="Detected Q-Peaks",
                q_peak_color=cmaps.med_dark[0],
                **kwargs,
            )
    if isinstance(algorithm, BaseBPointExtraction):
        c_point_algo = CPointExtractionScipyFindPeaks()
        c_points = c_point_algo.extract(
            icg=icg_data, heartbeats=heartbeats, sampling_rate_hz=datapoint.sampling_rate_icg
        ).points_
        algorithm.extract(
            icg=icg_data,
            heartbeats=heartbeats,
            c_points=c_points,
            sampling_rate_hz=datapoint.sampling_rate_icg,
        )
        b_points = algorithm.points_["b_point_sample"]
        b_points = b_points.loc[heartbeats.index].dropna()
        if collapse:
            _add_icg_b_points(
                icg_data,
                b_points,
                axs,
                b_point_label="Detected B-Points",
                b_point_color=cmaps.phil_dark[0],
                **kwargs,
            )
        else:
            _add_icg_b_points(
                icg_data,
                b_points,
                axs[1],
                b_point_label="Detected B-Points",
                b_point_color=cmaps.phil_dark[0],
                **kwargs,
            )

    if collapse:
        _handle_legend_one_axis(fig, axs, **kwargs)
    else:
        _handle_legend_two_axes(fig, axs, **kwargs)

    if not collapse:
        fig.align_ylabels()
    fig.tight_layout(rect=rect)
    return fig, axs


def plot_signals_from_challenge_results(
    datapoint: BasePepDatasetWithAnnotations,
    pep_results_per_sample: pd.DataFrame,
    *,
    collapse: bool = False,
    normalize_time: bool = False,
    heartbeat_subset: Sequence[int] | None = None,
    add_pep: bool = False,
    **kwargs: Any,
) -> tuple[plt.Figure, Sequence[plt.Axes]]:
    kwargs.setdefault("legend_loc", _get_legend_loc(kwargs))
    kwargs.setdefault("legend_max_cols", 5)
    rect = _get_rect(kwargs)

    fig, axs = plot_signals(
        datapoint, normalize_time=normalize_time, heartbeat_subset=heartbeat_subset, collapse=collapse, **kwargs
    )

    ecg_data, icg_data = _get_data(datapoint, normalize_time=normalize_time, heartbeat_subset=heartbeat_subset)

    labels_from_challenge = _get_labels_from_challenge_results(pep_results_per_sample, heartbeat_subset)

    print(ecg_data)
    print(labels_from_challenge["heartbeats_start"])

    heartbeats_start = ecg_data.index[labels_from_challenge["heartbeats_start"]]
    heartbeats_end = ecg_data.index[labels_from_challenge["heartbeats_end"] - 1]
    q_peak_labels_reference = labels_from_challenge["q_peak_labels_reference"]
    q_peak_labels_estimated = labels_from_challenge["q_peak_labels_estimated"]
    b_point_labels_reference = labels_from_challenge["b_point_labels_reference"]
    b_point_labels_estimated = labels_from_challenge["b_point_labels_estimated"]

    labels_reference = pd.concat({"ecg": q_peak_labels_reference, "icg": b_point_labels_reference}, axis=1)
    labels_estimated = pd.concat({"ecg": q_peak_labels_estimated, "icg": b_point_labels_estimated}, axis=1)

    if collapse:
        ax_ecg = axs
        ax_icg = axs
    else:
        ax_ecg = axs[0]
        ax_icg = axs[1]

    _add_heartbeat_borders(heartbeats_start, ax_ecg)
    _add_heartbeat_borders(heartbeats_end, ax_ecg)
    if not collapse:
        _add_heartbeat_borders(heartbeats_start, ax_icg)
        _add_heartbeat_borders(heartbeats_end, ax_icg)

    _add_ecg_q_peaks(
        ecg_data,
        q_peak_labels_reference,
        ax_ecg,
        q_peak_color=cmaps.med[0],
        q_peak_label="Q-Peak Reference",
        plot_artifacts=False,
    )
    _add_ecg_q_peaks(
        ecg_data,
        q_peak_labels_estimated,
        ax_ecg,
        q_peak_color=cmaps.med_dark[0],
        q_peak_label="Q-Peak Estimated",
        plot_artifacts=False,
    )

    _add_icg_b_points(
        icg_data,
        b_point_labels_reference,
        ax_icg,
        b_point_color=cmaps.phil[0],
        b_point_label="B-Point Reference",
        plot_artifacts=False,
    )
    _add_icg_b_points(
        icg_data,
        b_point_labels_estimated,
        ax_icg,
        b_point_color=cmaps.phil_dark[0],
        b_point_label="B-Point Estimated",
        plot_artifacts=False,
    )

    if add_pep:
        _add_pep_from_results(
            ecg_data,
            icg_data,
            labels_reference,
            ax=ax_icg,
            pep_color=cmaps.nat[0],
            pep_hatch="////",
            pep_label="PEP Reference",
        )
        _add_pep_from_results(
            ecg_data,
            icg_data,
            labels_estimated,
            ax=ax_icg,
            pep_color=cmaps.nat_dark[0],
            pep_hatch=r"\\\\",
            pep_label="PEP Estimated",
        )
        if not collapse:
            _add_pep_from_results(
                ecg_data,
                icg_data,
                labels_reference,
                ax=ax_ecg,
                pep_color=cmaps.nat[0],
                pep_hatch="////",
                pep_label="PEP Reference",
            )
            _add_pep_from_results(
                ecg_data,
                icg_data,
                labels_estimated,
                ax=ax_ecg,
                pep_color=cmaps.nat_dark[0],
                pep_hatch=r"\\\\",
                pep_label="PEP Estimated",
            )

    if collapse:
        _handle_legend_one_axis(fig, axs, **kwargs)
    else:
        _handle_legend_two_axes(fig, axs, **kwargs)

    fig.tight_layout(rect=rect)

    return fig, axs


def _plot_signals_one_axis(
    *,
    datapoint: BasePepDataset | None = None,
    df: pd.DataFrame | None = None,
    normalize_time: bool = False,
    heartbeat_subset: Sequence[int] | None = None,
    **kwargs: Any,
) -> tuple[plt.Figure, plt.Axes]:
    kwargs.setdefault("legend_loc", _get_legend_loc(kwargs))
    kwargs.setdefault("legend_max_cols", 5)
    plot_ecg = kwargs.get("plot_ecg", True)
    plot_icg = kwargs.get("plot_icg", True)
    color = kwargs.get("color", cmaps.fau[0])

    if datapoint is not None and df is not None:
        raise ValueError("Either `datapoint` or `df` must be provided, but not both.")
    if datapoint is None and df is None:
        raise ValueError("Either `datapoint` or `df` must be provided.")

    rect = _get_rect(kwargs)

    fig, ax = _get_fig_ax(kwargs)
    kwargs.pop("ax", None)

    x_label = "Time [hh:mm:ss]"

    if datapoint is not None:
        ecg_data, icg_data = _get_data(datapoint, normalize_time=normalize_time, heartbeat_subset=heartbeat_subset)

        if plot_ecg:
            ecg_data.columns = ["ECG"]
            ecg_data.plot(ax=ax)
            ax.legend()
        if plot_icg:
            icg_data.columns = ["ICG ($dZ/dt$)"]
            icg_data.plot(ax=ax)

        if normalize_time or not all(isinstance(data.index, pd.DatetimeIndex) for data in [ecg_data, icg_data]):
            x_label = "Time [s]"
    else:
        if isinstance(df, pd.Series):
            df = df.to_frame()
        df.columns = kwargs.get("columns", df.columns)
        df.plot(ax=ax, color=color)
        if normalize_time or not isinstance(df.index, pd.DatetimeIndex):
            x_label = "Time [s]"

    ax.set_xlabel(x_label)
    ax.set_ylabel("Amplitude [a.u.]")

    _handle_legend_one_axis(fig, ax, **kwargs)

    if kwargs.get("use_tight", True):
        fig.tight_layout(rect=rect)

    return fig, ax


def _plot_signals_two_axes(
    *,
    datapoint: BasePepDataset,
    normalize_time: bool | None = False,
    heartbeat_subset: Sequence[int] | None = None,
    **kwargs: Any,
) -> tuple[plt.Figure, Sequence[plt.Axes]]:
    kwargs.setdefault("nrows", 2)
    kwargs.setdefault("legend_loc", _get_legend_loc(kwargs))
    kwargs.setdefault("legend_max_cols", 5)

    rect = kwargs.get("rect", _get_rect(kwargs))

    fig, axs = _get_fig_axs(kwargs)

    colors = iter(cmaps.faculties)

    ecg_data, icg_data = _get_data(datapoint, normalize_time=normalize_time, heartbeat_subset=heartbeat_subset)
    ecg_data.columns = ["ECG"]
    icg_data.columns = ["ICG ($dZ/dt$)"]

    ecg_data.plot(ax=axs[0], color=next(colors), title="Electrocardiogram (ECG)")
    icg_data.plot(ax=axs[1], color=next(colors), title="Impedance Cardiogram (ICG)")

    _handle_legend_two_axes(fig, axs, **kwargs)

    for ax in axs:
        if normalize_time or not all(isinstance(data.index, pd.DatetimeIndex) for data in [ecg_data, icg_data]):
            ax.set_xlabel("Time [s]")
        else:
            ax.set_xlabel("Time [hh:mm:ss]")
        ax.set_ylabel("Amplitude [a.u.]")

    fig.align_ylabels()
    fig.tight_layout(rect=rect)

    return fig, axs


def _plot_blandaltman(  # noqa: PLR0915
    x: pd.Series | np.ndarray | list,
    y: pd.Series | np.ndarray | list,
    agreement: float = 1.96,
    xaxis: str = "mean",
    confidence: float = 0.95,
    annotate: bool = True,
    ax: plt.Axes | None = None,
    **kwargs: Any,
) -> plt.Axes:
    """
    Generate a Bland-Altman plot to compare two sets of measurements.

    Parameters
    ----------
    x, y : pd.Series, np.array, or list
        First and second measurements.
    agreement : float
        Multiple of the standard deviation to plot agreement limits.
        The defaults is 1.96, which corresponds to 95% confidence interval if
        the differences are normally distributed.
    xaxis : str
        Define which measurements should be used as the reference (x-axis).
        Default is to use the average of x and y ("mean"). Accepted values are
        "mean", "x" or "y".
    confidence : float
        If not None, plot the specified percentage confidence interval of
        the mean and limits of agreement. The CIs of the mean difference and
        agreement limits describe a possible error in the
        estimate due to a sampling error. The greater the sample size,
        the narrower the CIs will be.
    annotate : bool
        If True (default), annotate the values for the mean difference
        and agreement limits.
    ax : matplotlib axes
        Axis on which to draw the plot.
    **kwargs : optional
        Optional argument(s) passed to :py:func:`matplotlib.pyplot.scatter`.

    Returns
    -------
    ax : Matplotlib Axes instance
        Returns the Axes object with the plot for further tweaking.

    Notes
    -----
    Bland-Altman plots [1]_ are extensively used to evaluate the agreement
    among two different instruments or two measurements techniques.
    They allow identification of any systematic difference between the
    measurements (i.e., fixed bias) or possible outliers.

    The mean difference (= x - y) is the estimated bias, and the SD of the
    differences measures the random fluctuations around this mean.
    If the mean value of the difference differs significantly from 0 on the
    basis of a 1-sample t-test, this indicates the presence of fixed bias.
    If there is a consistent bias, it can be adjusted for by subtracting the
    mean difference from the new method.

    It is common to compute 95% limits of agreement for each comparison
    (average difference ± 1.96 standard deviation of the difference), which
    tells us how far apart measurements by 2 methods were more likely to be
    for most individuals. If the differences within mean ± 1.96 SD are not
    clinically important, the two methods may be used interchangeably.
    The 95% limits of agreement can be unreliable estimates of the population
    parameters especially for small sample sizes so, when comparing methods
    or assessing repeatability, it is important to calculate confidence
    intervals for the 95% limits of agreement.

    The code is an adaptation of the
    `PyCompare <https://github.com/jaketmp/pyCompare>`_ package. The present
    implementation is a simplified version; please refer to the original
    package for more advanced functionalities.

    References
    ----------
    .. [1] Bland, J. M., & Altman, D. (1986). Statistical methods for assessing
           agreement between two methods of clinical measurement. The lancet,
           327(8476), 307-310.

    .. [2] Giavarina, D. (2015). Understanding bland altman analysis.
           Biochemia medica, 25(2), 141-151.

    Examples
    --------
    Bland-Altman plot (example data from [2]_)
    """
    # Safety check
    assert xaxis in ["mean", "x", "y"]
    # Get names before converting to NumPy array
    xname = x.name if isinstance(x, pd.Series) else "x"
    yname = y.name if isinstance(y, pd.Series) else "y"
    x = np.asarray(x)
    y = np.asarray(y)
    assert x.ndim == 1
    assert y.ndim == 1
    assert x.size == y.size
    assert not np.isnan(x).any(), "Missing values in x or y are not supported."
    assert not np.isnan(y).any(), "Missing values in x or y are not supported."

    _annotate_kwargs = {key: val for key, val in kwargs.items() if key.startswith("annotate")}
    # remove annotate kwargs from kwargs
    kwargs = {key: val for key, val in kwargs.items() if not key.startswith("annotate")}

    # Update default kwargs with specified inputs
    _scatter_kwargs = {"color": "tab:blue", "alpha": 0.8}
    _scatter_kwargs.update(kwargs)

    # Calculate mean, STD and SEM of x - y
    n = x.size
    dof = n - 1
    diff = x - y
    mean_diff = np.mean(diff)
    std_diff = np.std(diff, ddof=1)
    mean_diff_se = np.sqrt(std_diff**2 / n)
    # Limits of agreements
    high = mean_diff + agreement * std_diff
    low = mean_diff - agreement * std_diff
    high_low_se = np.sqrt(3 * std_diff**2 / n)

    # Define x-axis
    if xaxis == "mean":
        xval = np.vstack((x, y)).mean(0)
        xlabel = f"Mean of {xname} and {yname}"
    elif xaxis == "x":
        xval = x
        xlabel = xname
    else:
        xval = y
        xlabel = yname

    # Start the plot
    if ax is None:
        ax = plt.gca()

    # Plot the mean diff, limits of agreement and scatter
    ax.scatter(xval, diff, **_scatter_kwargs)
    ax.axhline(mean_diff, color="k", linestyle="-", lw=2)
    ax.axhline(high, color="k", linestyle=":", lw=1.5)
    ax.axhline(low, color="k", linestyle=":", lw=1.5)

    # Annotate values
    if annotate:
        loa_range = high - low
        offset = (loa_range / 100.0) * 1.5
        trans = transforms.blended_transform_factory(ax.transAxes, ax.transData)
        annotate_fontsize = _annotate_kwargs.get("annotate_fontsize", "medium")
        annotate_bbox = _annotate_kwargs.get("annotate_bbox", False)

        bbox = transforms.Bbox.null()
        xloc = 0.98
        t = ax.text(
            xloc,
            mean_diff + offset,
            "Mean",
            ha="right",
            va="bottom",
            transform=trans,
            # bbox=bbox,
            fontdict={"fontsize": annotate_fontsize},
        )
        bbox.update_from_data_xy(_get_bbox_coords(t, ax), ignore=False)

        t = ax.text(
            xloc,
            mean_diff - offset,
            f"{mean_diff:.2f}",
            ha="right",
            va="top",
            transform=trans,
            # bbox=bbox,
            fontdict={"fontsize": annotate_fontsize},
        )
        bbox.update_from_data_xy(_get_bbox_coords(t, ax), ignore=False)

        t = ax.text(
            xloc,
            high + offset,
            f"+{agreement:.2f} SD",
            ha="right",
            va="bottom",
            transform=trans,
            # bbox=bbox,
            fontdict={"fontsize": annotate_fontsize},
        )
        bbox.update_from_data_xy(_get_bbox_coords(t, ax), ignore=False)

        t = ax.text(
            xloc,
            high - offset,
            f"{high:.2f}",
            ha="right",
            va="top",
            transform=trans,
            # bbox=bbox,
            fontdict={"fontsize": annotate_fontsize},
        )
        bbox.update_from_data_xy(_get_bbox_coords(t, ax), ignore=False)

        t = ax.text(
            xloc,
            low + offset,
            f"{low:.2f}",
            ha="right",
            va="bottom",
            transform=trans,
            # bbox=bbox,
            fontdict={"fontsize": annotate_fontsize},
        )
        bbox.update_from_data_xy(_get_bbox_coords(t, ax), ignore=False)

        t = ax.text(
            xloc,
            low - offset,
            f"-{agreement:.2f} SD",
            ha="right",
            va="top",
            transform=trans,
            # bbox=bbox,
            fontsize=annotate_fontsize,
        )
        bbox.update_from_data_xy(_get_bbox_coords(t, ax), ignore=False)
        if annotate_bbox:
            add_fancy_patch_around(ax, bbox)

    # Add 95% confidence intervals for mean bias and limits of agreement
    if confidence is not None:
        assert 0 < confidence < 1
        ci = {
            "mean": stats.t.interval(confidence, dof, loc=mean_diff, scale=mean_diff_se),
            "high": stats.t.interval(confidence, dof, loc=high, scale=high_low_se),
            "low": stats.t.interval(confidence, dof, loc=low, scale=high_low_se),
        }
        ax.axhspan(ci["mean"][0], ci["mean"][1], facecolor="tab:grey", alpha=0.2)
        ax.axhspan(ci["high"][0], ci["high"][1], facecolor=_scatter_kwargs["color"], alpha=0.2)
        ax.axhspan(ci["low"][0], ci["low"][1], facecolor=_scatter_kwargs["color"], alpha=0.2)

    # Labels
    ax.set_ylabel(f"{xname} - {yname}")
    ax.set_xlabel(xlabel)
    sns.despine(ax=ax)
    return ax


def _plot_paired(  # noqa: PLR0915, PLR0912, C901
    data: pd.DataFrame,
    dv: str,
    within: str,
    subject: str,
    order: Sequence[str] | None = None,
    boxplot: bool = True,
    boxplot_in_front: bool = False,
    orient: str = "v",
    ax: plt.Axes | None = None,
    colors: Sequence[str] | None = None,
    pointplot_kwargs: dict | None = None,
    boxplot_kwargs: dict | None = None,
) -> tuple[plt.Figure, plt.Axes]:
    """
    Paired plot.

    Parameters
    ----------
    data : :py:class:`pandas.DataFrame`
        Long-format dataFrame.
    dv : string
        Name of column containing the dependent variable.
    within : string
        Name of column containing the within-subject factor.
    subject : string
        Name of column containing the subject identifier.
    order : list of str
        List of values in ``within`` that define the order of elements on the
        x-axis of the plot. If None, uses alphabetical order.
    boxplot : boolean
        If True, add a boxplot to the paired lines using the
        :py:func:`seaborn.boxplot` function.
    boxplot_in_front : boolean
        If True, the boxplot is plotted on the foreground (i.e. above the
        individual lines) and with a slight transparency. This makes the
        overall plot more readable when plotting a large numbers of subjects.

        .. versionadded:: 0.3.8
    orient : string
        Plot the boxplots vertically and the subjects on the x-axis if
        ``orient='v'`` (default). Set to ``orient='h'`` to rotate the plot by
        by 90 degrees.

        .. versionadded:: 0.3.9
    ax : matplotlib axes
        Axis on which to draw the plot.
    colors : list of str
        Line colors names. Default is green when value increases from A to B,
        indianred when value decreases from A to B and grey when the value is
        the same in both measurements.
    pointplot_kwargs : dict
        Dictionnary of optional arguments that are passed to the
        :py:func:`seaborn.pointplot` function.
    boxplot_kwargs : dict
        Dictionnary of optional arguments that are passed to the
        :py:func:`seaborn.boxplot` function.

    Returns
    -------
    ax : Matplotlib Axes instance
        Returns the Axes object with the plot for further tweaking.

    Notes
    -----
    Data must be a long-format pandas DataFrame. Missing values are automatically removed using a
    strict listwise approach (= complete-case analysis).

    Examples
    --------
    Default paired plot:

    .. plot::

        >>> import pingouin as pg
        >>> df = pg.read_dataset("mixed_anova").query("Time != 'January'")
        >>> df = df.query("Group == 'Meditation' and Subject > 40")
        >>> fig, ax = pg._plot_paired(data=df, dv="Scores", within="Time", subject="Subject")

    Paired plot on an existing axis (no boxplot and uniform color):

    .. plot::

        >>> import pingouin as pg
        >>> import matplotlib.pyplot as plt
        >>> df = pg.read_dataset("mixed_anova").query("Time != 'January'")
        >>> df = df.query("Group == 'Meditation' and Subject > 40")
        >>> fig, ax1 = plt.subplots(1, 1, figsize=(5, 4))
        >>> pg._plot_paired(
        ...     data=df,
        ...     dv="Scores",
        ...     within="Time",
        ...     subject="Subject",
        ...     ax=ax1,
        ...     boxplot=False,
        ...     colors=["grey", "grey", "grey"],
        ... )  # doctest: +SKIP

    Horizontal paired plot with three unique within-levels:

    .. plot::

        >>> import pingouin as pg
        >>> import matplotlib.pyplot as plt
        >>> df = pg.read_dataset("mixed_anova").query("Group == 'Meditation'")
        >>> # df = df.query("Group == 'Meditation' and Subject > 40")
        >>> fig, ax = pg._plot_paired(
        >>>     data=df, dv="Scores", within="Time", subject="Subject", orient="h"
        >>> )  # doctest: +SKIP

    With the boxplot on the foreground:

    .. plot::

        >>> import pingouin as pg
        >>> df = pg.read_dataset("mixed_anova").query("Time != 'January'")
        >>> df = df.query("Group == 'Control'")
        >>> fig, ax = pg._plot_paired(data=df, dv="Scores", within="Time", subject="Subject", boxplot_in_front=True)
    """
    from pingouin.utils import _check_dataframe

    # Set default colors
    if colors is None:
        colors = ["green", "grey", "indianred"]

    if pointplot_kwargs is None:
        pointplot_kwargs = {}
    if boxplot_kwargs is None:
        boxplot_kwargs = {}

    # Update default kwargs with specified inputs
    _pointplot_kwargs = {"scale": 0.6, "marker": "."}
    _pointplot_kwargs.update(pointplot_kwargs)
    _boxplot_kwargs = {"color": "lightslategrey", "width": 0.2}
    _boxplot_kwargs.update(boxplot_kwargs)
    # Extract pointplot alpha, if set
    pp_alpha = _pointplot_kwargs.pop("alpha", 1.0)

    # Calculate size of the plot elements by scale as in Seaborn pointplot
    scale = _pointplot_kwargs.pop("scale")
    lw = plt.rcParams["lines.linewidth"] * 1.8 * scale  # get the linewidth
    mew = lw * 0.75  # get the markeredgewidth
    markersize = np.pi * np.square(lw) * 2  # get the markersize

    # Set boxplot in front of Line2D plot (zorder=2 for both) and add alpha
    if boxplot_in_front:
        _boxplot_kwargs.update(
            {
                "boxprops": {"zorder": 2},
                "whiskerprops": {"zorder": 2},
                "zorder": 2,
            }
        )

    # Validate args
    data = _check_dataframe(data=data, dv=dv, within=within, subject=subject, effects="within")

    # Pivot and melt the table. This has several effects:
    # 1) Force missing values to be explicit (a NaN cell is created)
    # 2) Automatic collapsing to the mean if multiple within factors are present
    # 3) If using dropna, remove rows with missing values (listwise deletion).
    # The latter is the same behavior as JASP (= strict complete-case analysis).
    data_piv = data.pivot_table(index=subject, columns=within, values=dv, observed=True)
    data_piv = data_piv.dropna()
    data = data_piv.melt(ignore_index=False, value_name=dv).reset_index()

    # Extract within-subject level (alphabetical order)
    x_cat = np.unique(data[within])

    if order is None:
        order = x_cat
    else:
        assert len(order) == len(x_cat), (
            "Order must have the same number of elements as the number of levels in `within`."
        )

    # Substitute within by integer order of the ordered columns to allow for
    # changing the order of numeric withins.
    data["wthn"] = data[within].replace({_ordr: str(i) for i, _ordr in enumerate(order)})
    data["wthn"] = data["wthn"].astype(int)
    order_num = range(len(order))  # Make numeric order

    # Start the plot
    if ax is None:
        fig, ax = _get_fig_ax({})
    else:
        fig = ax.get_figure()

    # Set x and y depending on orientation using the num. replacement within
    _x = "wthn" if orient == "v" else dv
    _y = dv if orient == "v" else "wthn"

    for cat in range(len(x_cat) - 1):
        _order = (order_num[cat], order_num[cat + 1])
        # Extract data of the current subject-combination
        data_now = data.loc[data["wthn"].isin(_order), [dv, "wthn", subject]]
        # Select colors for all lines between the current subjects
        y1 = data_now.loc[data_now["wthn"] == _order[0], dv].to_numpy()
        y2 = data_now.loc[data_now["wthn"] == _order[1], dv].to_numpy()
        # Line and scatter colors depending on subject dv trend
        _colors = np.where(y1 < y2, colors[0], np.where(y1 > y2, colors[2], colors[1]))
        # Line and scatter colors as hue-indexed dictionary
        _colors = dict(zip(data_now[subject].unique(), _colors, strict=False))
        # Plot individual lines using Seaborn
        sns.lineplot(
            data=data_now,
            x=_x,
            y=_y,
            hue=subject,
            palette=_colors,
            ls="-",
            lw=lw,
            legend=False,
            ax=ax,
        )
        # Plot individual markers using Seaborn
        sns.scatterplot(
            data=data_now,
            x=_x,
            y=_y,
            hue=subject,
            palette=_colors,
            edgecolor="face",
            lw=mew,
            sizes=[markersize] * data_now.shape[0],
            legend=False,
            ax=ax,
            **_pointplot_kwargs,
        )

    # Set zorder and alpha of pointplot markers and lines
    _ = plt.setp(ax.collections, alpha=pp_alpha, zorder=2)  # Set marker alpha
    _ = plt.setp(ax.lines, alpha=pp_alpha, zorder=2)  # Set line alpha

    if boxplot:
        # Set boxplot x and y depending on orientation
        _xbp = within if orient == "v" else dv
        _ybp = dv if orient == "v" else within
        sns.boxplot(data=data, x=_xbp, y=_ybp, order=order, ax=ax, orient=orient, **_boxplot_kwargs)

        # Set alpha to patch of boxplot but not to whiskers
        for patch in ax.artists:
            r, g, b, a = patch.get_facecolor()
            patch.set_facecolor((r, g, b, 0.75))
    else:
        # If no boxplot, axis needs manual styling as in Seaborn pointplot
        if orient == "v":
            xlabel, ylabel = within, dv
            ax.set_xticks(np.arange(len(x_cat)))
            ax.set_xticklabels(order)
            ax.xaxis.grid(False)
            ax.set_xlim(-0.5, len(x_cat) - 0.5, auto=None)
        else:
            xlabel, ylabel = dv, within
            ax.set_yticks(np.arange(len(x_cat)))
            ax.set_yticklabels(order)
            ax.yaxis.grid(False)
            ax.set_ylim(-0.5, len(x_cat) - 0.5, auto=None)
            ax.invert_yaxis()
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

    # Despine and trim
    sns.despine(trim=True, ax=ax)
    return fig, ax



from typing import get_args

from biopsykit.signals._base_extraction import CanHandleMissingEventsMixin
from biopsykit.signals.pep._pep_extraction import NEGATIVE_PEP_HANDLING
from typing_extensions import Self

__all__ = ["PepExtractionPipeline"]

from pepbench.pipelines._base_pipeline import BasePepDatasetT, BasePepExtractionPipeline, base_pep_pipeline_docfiller


@base_pep_pipeline_docfiller
class PepExtractionPipeline(BasePepExtractionPipeline):
    """Standard `tpcp` Pipeline for PEP extraction from ECG and ICG data.

    This pipeline can be used to run the PEP extraction algorithms on a given dataset. It will use the algorithms
    passed to the ``__init__`` method to extract PEP from the data. The results will be stored in the attributes of
    the class.


    %(base_parameters)s

    Other Parameters
    ----------------
    %(datapoint_pipeline)s

    %(attributes)s

    """

    @base_pep_pipeline_docfiller
    def run(self, datapoint: BasePepDatasetT) -> Self:
        """Run the pipeline on the given datapoint.

        The pipeline will extract PEP from the given datapoint using the specified algorithms. The results will be
        stored in the attributes of the class.

        Parameters
        ----------
        %(datapoint_pipeline)s

        Returns
        -------
        Self
            The instance of the class with the results set.

        Raises
        ------
        ValueError
            If the `handle_negative_pep` parameter is not one of the allowed values.

        """
        if self.handle_negative_pep not in get_args(NEGATIVE_PEP_HANDLING):
            raise ValueError(
                f"Invalid value for 'handle_negative_pep': {self.handle_negative_pep}. "
                f"Must be one of {NEGATIVE_PEP_HANDLING}"
            )

        heartbeat_algo = self.heartbeat_segmentation_algo.clone()
        q_peak_algo = self.q_peak_algo.clone()
        c_point_algo = self.c_point_algo.clone()
        b_point_algo = self.b_point_algo.clone()
        outlier_algo = self.outlier_correction_algo.clone()

        fs_ecg = datapoint.sampling_rate_ecg
        fs_icg = datapoint.sampling_rate_icg

        ecg_data = datapoint.ecg
        icg_data = datapoint.icg

        # set handle_missing parameter for all algorithms
        if self.handle_missing_events is not None:
            for algo in (heartbeat_algo, q_peak_algo, c_point_algo, b_point_algo, outlier_algo):
                if isinstance(algo, CanHandleMissingEventsMixin):
                    # this overwrites the default value of the handle_missing parameter
                    algo.set_params(handle_missing_events=self.handle_missing_events)

        # extract heartbeats
        heartbeat_algo.extract(ecg=ecg_data, sampling_rate_hz=fs_ecg)
        heartbeats = heartbeat_algo.heartbeat_list_

        # run Q-peak extraction
        q_peak_algo.extract(ecg=ecg_data, heartbeats=heartbeats, sampling_rate_hz=fs_ecg)
        q_peak_samples = q_peak_algo.points_

        # run C-point extraction
        c_point_algo.extract(icg=icg_data, heartbeats=heartbeats, sampling_rate_hz=fs_icg)

        # run B-point extraction
        b_point_algo.extract(
            icg=icg_data,
            heartbeats=heartbeats,
            c_points=c_point_algo.points_,
            sampling_rate_hz=fs_icg,
        )

        outlier_algo.correct_outlier(
            b_points=b_point_algo.points_,
            c_points=c_point_algo.points_,
            sampling_rate_hz=fs_icg,
            handle_negative=self.handle_negative_pep,
        )
        b_point_samples_after_outlier = outlier_algo.points_

        pep_results = self._compute_pep(
            heartbeats=heartbeats,
            q_peak_samples=q_peak_samples,
            b_point_samples=b_point_samples_after_outlier,
            sampling_rate_hz=fs_icg,
        )

        self.heartbeat_segmentation_results_ = heartbeats
        self.q_peak_results_ = q_peak_samples
        self.c_point_results_ = c_point_algo.points_
        self.b_point_results_ = b_point_algo.points_
        self.b_point_after_outlier_correction_results_ = b_point_samples_after_outlier
        self.pep_results_ = pep_results
        return self



from typing import ClassVar

import numpy as np
import pandas as pd
from biopsykit.utils.array_handling import downsample
from scipy.io import loadmat

from pepbench.utils._types import path_t


class TFMLoader:
    """Class for loading and processing Task Force Monitor (TFM) data from the Guardian Study."""

    SAMPLING_RATES_IN: ClassVar[dict[str, float]] = {"ecg_1": 1000, "ecg_2": 1000, "icg_der": 500}
    SAMPLING_RATE_OUT = 500
    CHANNEL_MAPPING: ClassVar[dict[str, str]] = {"ecg_1": "rawECG1", "ecg_2": "rawECG2", "icg_der": "rawICG"}
    PHASES: ClassVar[list[str]] = [
        "BeginRecording",
        "Pause",
        "Valsalva",
        "HoldingBreath",
        "TiltUp",
        "TiltDown",
    ]

    ORIGINAL_NAMES: ClassVar[dict[str, str]] = {
        "BeginRecording": "Beginn der Aufzeichnung",
        "Pause": "Ruhe",
        "Valsalva": "Valsalva",
        "HoldingBreath": "AtmungAnhalten",
        "TiltUp": "TiltUp",
        "TiltDown": "TiltDown",
    }
    _start_time_unix: pd.Timestamp
    _tz: str
    _start_time_dict: dict[str, pd.Timestamp]

    def __init__(
        self,
        data_dict: dict[str, pd.DataFrame],
        sampling_rate_dict: dict[str, float],
        start_time_dict: dict[str, pd.Timestamp],
        tz: str | None = None,
    ) -> None:
        """Initialize a TFM dataset.

        Parameters
        ----------
        data_dict : dict
            Dictionary containing TFM data. Keys are channel names, values are dataframes with the TFM data.
        tz : str, optional
            Timezone of the data. Default: None

        """
        self._data = data_dict
        self._sampling_rate = sampling_rate_dict
        for name, data in data_dict.items():
            setattr(self, name, data)
        for name, sampling_rate in sampling_rate_dict.items():
            setattr(self, f"sampling_rate_hz_{name}", sampling_rate)
        setattr(self, "channels", list(self._data.keys()))
        self._start_time_dict = start_time_dict
        self._tz = tz

    @classmethod
    def from_mat_file(  # noqa: C901, PLR0912
        cls,
        file_path: path_t,
        date: str | pd.Timestamp,
        tz: str | None = "Europe/Berlin",
    ) -> "TFMLoader":
        data = loadmat(file_path, struct_as_record=False, squeeze_me=True)

        data_raw = data["RAW_SIGNALS"]
        names = data["IV"].Name
        times = data["IV"].AbsTime

        data_dict_tmp = {key: getattr(data_raw, value) for key, value in cls.CHANNEL_MAPPING.items()}

        data_dict = {}

        for key, value in data_dict_tmp.items():
            data_dict[key] = {}

            for key1, value1 in cls.ORIGINAL_NAMES.items():
                if value1 not in names:
                    data_dict[key][key1] = np.array([])
                else:
                    index = list(names).index(value1)
                    data_dict[key][key1] = value[index]

        for key, value in data_dict.items():
            if cls.SAMPLING_RATES_IN[key] != cls.SAMPLING_RATE_OUT:
                for recording_key, recording_value in value.items():
                    if isinstance(recording_value, np.ndarray):
                        if len(recording_value) == 0:
                            continue
                        data_dict[key][recording_key] = downsample(
                            recording_value,
                            fs_in=cls.SAMPLING_RATES_IN[key],
                            fs_out=cls.SAMPLING_RATE_OUT,
                        )

        new_data = {}

        for key, value in data_dict.items():
            for recording_key, recording_value in value.items():
                if recording_key not in new_data:
                    new_data[recording_key] = {}

                new_data[recording_key][key] = recording_value

        for key2, _value2 in new_data.items():
            new_data[key2] = pd.concat(
                [
                    pd.Series(new_data[key2]["ecg_1"]),
                    pd.Series(new_data[key2]["ecg_2"]),
                    pd.Series(new_data[key2]["icg_der"]),
                ],
                axis=1,
            )
            new_data[key2].columns = cls.CHANNEL_MAPPING.keys()
            new_data[key2] = new_data[key2].dropna(axis=0)

        start_time_dict = {}

        for key1, value1 in cls.ORIGINAL_NAMES.items():
            if value1 not in names:
                start_time_dict[key1] = pd.Timestamp("2000-01-01 00:00:00", tz="Europe/Berlin")
            else:
                index = list(names).index(value1)
                start_time_dict[key1] = date.tz_localize(tz) + pd.Timedelta(times[index])

        return cls(
            data_dict=new_data,
            tz=tz,
            sampling_rate_dict={},
            start_time_dict=start_time_dict,
        )

    @property
    def start_time_unix(self) -> pd.Timestamp | None:
        """Start time of the recording in UTC time."""
        return self._start_time_unix

    @property
    def timezone(self) -> str:
        """Timezone the dataset was recorded in."""
        return self._tz

    def data_as_dict(self, index: str | None = None) -> dict[str, pd.DataFrame]:
        return {
            key: self._add_index(val, index=index, start_time=start_time)
            for (key, val), start_time in zip(self._data.items(), self._start_time_dict.values(), strict=False)
        }

    def _add_index(self, data: pd.DataFrame, index: str, start_time: pd.Timestamp | None = None) -> pd.DataFrame:
        index_names = {
            None: "n_samples",
            "time": "t",
            "utc": "utc",
            "utc_datetime": "date",
            "local_datetime": f"date ({self.timezone})",
        }
        if index and index not in index_names:
            raise ValueError(f"Supplied value for index ({index}) is not allowed. Allowed values: {index_names.keys()}")
        index_name = index_names[index]
        data.index.name = index_name

        if index is None:
            data = data.reset_index(drop=True)
            data.index.name = index_name
            return data

        data.index = data.index / self.SAMPLING_RATE_OUT

        if index == "time":
            return data

        if index == "utc":
            # convert counter to utc timestamps
            time_object = pd.to_datetime(start_time)
            data.index += time_object.timestamp()
            return data

        if start_time is None:
            start_time = self.start_time_unix

        if start_time is None:
            raise ValueError(
                "No start time available - can't convert to datetime index! "
                "Use a different index representation or provide a custom start time using the 'start_time' parameter."
            )

        # convert counter to pandas datetime index
        data.index = pd.to_timedelta(data.index, unit="s")
        start_time = pd.to_datetime(start_time)
        data.index += start_time
        data.index = pd.to_datetime(data.index, unit="s").tz_convert("UTC")

        if index == "local_datetime":
            data.index = pd.to_datetime(data.index, unit="s").tz_convert(self.timezone)

        return data



from collections.abc import Sequence
from functools import lru_cache
from itertools import product
from pathlib import Path
from typing import ClassVar

import pandas as pd
from biopsykit.metadata import bmi
from biopsykit.signals.ecg.preprocessing import EcgPreprocessingNeurokit
from biopsykit.signals.ecg.segmentation import HeartbeatSegmentationNeurokit
from biopsykit.signals.icg.preprocessing import IcgPreprocessingBandpass

from pepbench.datasets import BasePepDatasetWithAnnotations
from pepbench.datasets._base_pep_extraction_dataset import MetadataMixin, base_pep_extraction_docfiller
from pepbench.datasets._helper import compute_reference_heartbeats, load_labeling_borders
from pepbench.datasets.guardian._helper import _load_tfm_data
from pepbench.utils._types import path_t

__all__ = ["GuardianDataset"]


_cached_get_tfm_data = lru_cache(maxsize=4)(_load_tfm_data)
# cache_dir = "./cachedir"
# memory = Memory(location=cache_dir, verbose=0)
# _cached_get_tfm_data = memory.cache(_load_tfm_data)


@base_pep_extraction_docfiller
class GuardianDataset(BasePepDatasetWithAnnotations, MetadataMixin):
    """Dataset class for the Guardian Dataset.

    This class is the ``tpcp`` dataset class for the Guardian dataset. It provides access to the Task Force Monitor
    (TFM) data (for ECG and ICG), the reference annotations for the ECG and ICG annotations, as well as metadata
    like age, gender, and BMI.

    Attributes
    ----------
    %(base_attributes_pep)s
    %(base_attributes_pep_label)s
    %(base_attributes_metadata)s
    timelog : :class:`~pandas.DataFrame`
        Timelog data, indicating the start and end of each experimental phase, as a pandas DataFrame.

    """

    base_path: Path
    use_cache: bool

    SAMPLING_RATES: ClassVar[dict[str, int]] = {"ecg_1": 500, "ecg_2": 500, "icg_der": 500}
    PHASES: ClassVar[tuple[str, ...]] = ["Pause", "Valsalva", "HoldingBreath", "TiltUp", "TiltDown"]

    GENDER_MAPPING: ClassVar[dict[str, str]] = {"M": "Male", "F": "Female"}

    SUBSET_NO_RECORDED_DATA = (
        ("GDN0006", "HoldingBreath"),
        ("GDN0009", "HoldingBreath"),
        ("GDN0010", "Valsalva"),
        ("GDN0017", "Pause"),
        ("GDN0018", "TiltDown"),
        ("GDN0020", "TiltUp"),
        ("GDN0022", "TiltUp"),
        ("GDN0024", "TiltDown"),
        ("GDN0025", "Valsalva"),
        ("GDN0028", "TiltUp"),
        ("GDN0030", "Pause"),
        ("GDN0030", "TiltUp"),
    )
    SUBSET_NOISY_DATA = (
        ("GDN0025", "TiltUp"),
        ("GDN0025", "TiltDown"),
    )

    def __init__(
        self,
        base_path: path_t,
        groupby_cols: Sequence[str] | None = None,
        subset_index: Sequence[str] | None = None,
        *,
        return_clean: bool = True,
        exclude_no_recorded_data: bool = True,
        exclude_noisy_data: bool = True,
        use_cache: bool = True,
        only_labeled: bool = False,
        label_type: str = "rater_01",
    ) -> None:
        """Initialize a new ``GuardianDataset`` instance.

        Parameters
        ----------
        base_path : :class:`~pathlib.Path` or str
            Path to the root directory of the Guardian dataset.
        return_clean : bool
            Whether to return the preprocessed/cleaned ECG and ICG data when accessing the respective properties.
            Default: ``True``.
        exclude_no_recorded_data : bool, optional
            Whether to exclude participants with no recorded data. Default: ``True``.
        exclude_noisy_data : bool, optional
            Whether to exclude participants with noisy data. Default: ``True``.
        use_cache : bool, optional
            Whether to use caching for loading TFM data. Default: ``True``.
        only_labeled : bool, optional
            Whether to only return segments that are labeled (i.e., cut the data to the labeling borders).
            This is necessary when using the dataset for evaluating the performance of PEP extraction algorithms or for
            training ML-based PEP extraction algorithms. Default: ``False``.
        label_type: str, optional
            Which annotations to use. Can be either "rater_01", "rater_02", or "average". Default: "rater_01".
        """
        self.base_path = base_path
        self.exclude_no_recorded_data = exclude_no_recorded_data
        self.exclude_noisy_data = exclude_noisy_data
        self.data_to_exclude = self._find_data_to_exclude()
        self.use_cache = use_cache
        self.label_type = label_type
        super().__init__(
            groupby_cols=groupby_cols, subset_index=subset_index, return_clean=return_clean, only_labeled=only_labeled
        )

    def _sanitize_params(self) -> None:
        # ensure pathlib
        self.base_path = Path(self.base_path)

    def create_index(self) -> pd.DataFrame:
        self._sanitize_params()
        overview_df = pd.read_csv(self.base_path.joinpath("metadata/dataset_overview.csv"), sep=";")
        pids = list(overview_df["participant"])
        index = list(product(pids, self.PHASES))
        index = pd.DataFrame(index, columns=["participant", "phase"])
        for item in self.data_to_exclude:
            index = index.drop(index[(index["participant"] == item[0]) & (index["phase"] == item[1])].index)
        index = index.reset_index(drop=True)

        return index

    def _find_data_to_exclude(self) -> Sequence[tuple[str, str]]:
        data_to_exclude = []
        if self.exclude_no_recorded_data:
            data_to_exclude.extend(self.SUBSET_NO_RECORDED_DATA)
        if self.exclude_noisy_data:
            data_to_exclude.extend(self.SUBSET_NOISY_DATA)

        return data_to_exclude

    @property
    def sampling_rates(self) -> dict[str, int]:
        """Return the sampling rates of the ECG and ICG signals.

        Returns
        -------
        dict
            Dictionary with the sampling rates of the ECG and ICG signals in Hz.

        """
        return self.SAMPLING_RATES

    @property
    def sampling_rate_ecg(self) -> int:
        """Return the sampling rate of the ECG signal.

        Returns
        -------
        int
            Sampling rate of the ECG signal in Hz.

        """
        return self.SAMPLING_RATES["ecg_2"]

    @property
    def sampling_rate_icg(self) -> int:
        """Return the sampling rate of the ICG signal.

        Returns
        -------
        int
            Sampling rate of the ICG signal in Hz.

        """
        return self.SAMPLING_RATES["icg_der"]

    @property
    def date(self) -> pd.Series | pd.Timestamp:
        """Return the recording date of the participant(s).

        Returns
        -------
        pd.Series or pd.Timestamp
            Recording date of the participant(s). If only a single participant is selected, a single timestamp is
            returned. If multiple participants are selected, a :class:`~pandas.Series` with the recording dates for
            each participant is returned.

        """
        metadata_path = self.base_path.joinpath("metadata/recording_timestamps.xlsx")
        metadata = pd.read_excel(metadata_path)
        metadata = metadata.set_index("participant")["date"]
        if self.is_single("participant"):
            return metadata[self.index["participant"][0]]
        return metadata

    @property
    def tfm_data(self) -> pd.DataFrame | dict[str, pd.DataFrame]:
        """Return the Task Force Monitor (TFM) data.

        The data is loaded from the raw TFM data files and returned as :class:`~pandas.DataFrame`. The data can only be
        accessed for a single participant and a single phase or for a single participant and all phases.


        Returns
        -------
        :class:`~pandas.DataFrame` or dict
            TFM data as :class:`~pandas.DataFrame` or dictionary of :class:`~pandas.DataFrame` if multiple phases are
            selected.

        Raises
        ------
        ValueError
            If the data is accessed for multiple participants or multiple phases (that are not all phases).

        """
        participant = self.index["participant"][0]
        phases = self.index["phase"]

        if not (self.is_single(None) or len(phases) == len(self.PHASES)):
            raise ValueError(
                "TFM data can only be accessed for a single participant and ALL phases or "
                "for a single participant and a SINGLE phase."
            )

        tfm_path = self.base_path.joinpath(f"data_raw/{participant}/tfm_data/{participant.lower()}_no01.mat")

        if self.use_cache:
            tfm_data_dict = _cached_get_tfm_data(tfm_path, self.date)
        else:
            tfm_data_dict = _load_tfm_data(tfm_path, self.date)

        if self.is_single(None):
            tfm_data_dict = tfm_data_dict[phases[0]]

        if self.only_labeled:
            labeling_borders = self.labeling_borders

            if self.is_single(None):
                tfm_data_dict = self._cut_to_labeling_borders(tfm_data_dict, labeling_borders)
            else:
                for phase in phases:
                    borders = labeling_borders[labeling_borders["description"].apply(lambda x, ph=phase: ph in x)]
                    tfm_data = tfm_data_dict[phase]
                    tfm_data_dict[phase] = self._cut_to_labeling_borders(tfm_data, borders)

        return tfm_data_dict

    @property
    def icg(self) -> pd.DataFrame:
        """Return the ICG signal.

        If ``return_clean`` is set to ``True`` in the ``__init__``, the ICG signal is preprocessed and cleaned using
        the :class:`~biopsykit.signals.icg.preprocessing.IcgPreprocessingBandpass` algorithm before returning it.

        Returns
        -------
        :class:`~pandas.DataFrame`
            ICG signal as :class:`~pandas.DataFrame`.

        """
        if not self.is_single(None):
            raise ValueError("ICG data can only be accessed for a single participant and a single phase!")
        icg = self.tfm_data[["icg_der"]]
        if self.return_clean:
            algo = IcgPreprocessingBandpass()
            algo.clean(icg=icg, sampling_rate_hz=self.sampling_rate_icg)
            return algo.icg_clean_
        return icg

    @property
    def ecg(self) -> pd.DataFrame:
        """Return the ECG signal.

        If ``return_clean`` is set to ``True`` in the ``__init__``, the ECG signal is preprocessed and cleaned using the
        :class:`~biopsykit.signals.ecg.preprocessing.EcgPreprocessingNeurokit` algorithm before returning it.

        Returns
        -------
        :class:`~pandas.DataFrame`
            ECG signal as :class:`~pandas.DataFrame`.

        """
        if not self.is_single(None):
            raise ValueError("ECG data can only be accessed for a single participant and a single phase!")
        ecg = self.tfm_data[["ecg_2"]]
        ecg.columns = ["ecg"]

        if self.return_clean:
            algo = EcgPreprocessingNeurokit()
            algo.clean(ecg=ecg, sampling_rate_hz=self.sampling_rate_ecg)
            return algo.ecg_clean_
        return ecg

    @property
    def labeling_borders(self) -> pd.DataFrame:
        """Return the labeling borders for a selected participant and phase(s).

        Returns
        -------
        :class:`~pandas.DataFrame`
            Labeling borders as :class:`~pandas.DataFrame`.

        """
        participant = self.index["participant"][0]

        if not self.is_single("participant"):
            raise ValueError("Labeling borders can only be accessed for a single participant.")

        file_path = self.base_path.joinpath(
            f"data_raw/{participant}/tfm_data/reference_labels/labeling_borders_{participant}.csv"
        )
        data = load_labeling_borders(file_path)

        if self.is_single(None):
            phase = self.index["phase"][0]
            data = data[data["description"].apply(lambda x: phase in x)]

        return data

    @property
    def reference_heartbeats(self) -> pd.DataFrame:
        """Return the reference heartbeats.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference heartbeats as a pandas DataFrame

        """
        return self._load_reference_heartbeats()

    @property
    def reference_labels_ecg(self) -> pd.DataFrame:
        """Return the reference labels for the ECG signal.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference labels for the ECG signal as a pandas DataFrame

        """
        return self._load_reference_labels("ECG")

    @property
    def reference_labels_icg(self) -> pd.DataFrame:
        """Return the reference labels for the ICG signal.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference labels for the ICG signal as a pandas DataFrame

        """
        return self._load_reference_labels("ICG")

    def _load_reference_heartbeats(self) -> pd.DataFrame:
        reference_ecg = self.reference_labels_ecg
        reference_heartbeats = reference_ecg.reindex(["heartbeat"], level="channel")
        reference_heartbeats = compute_reference_heartbeats(reference_heartbeats)
        return reference_heartbeats

    def _load_reference_labels(self, channel: str) -> pd.DataFrame:
        participant = self.index["participant"][0]
        phases = self.index["phase"]

        if not (self.is_single(None) or len(phases) == len(self.PHASES)):
            raise ValueError(
                "Reference data can only be accessed for a single participant and ALL phases or "
                "for a single participant and a SINGLE phase."
            )
        reference_data_dict = {}
        rater_type = self.label_type
        if self.label_type == "average":
            # TODO implement
            raise NotImplementedError("Average reference labels are not implemented yet.")
        for phase in phases:
            file_path = self.base_path.joinpath(
                f"data_raw/{participant}/tfm_data/reference_labels/{rater_type}/"
                f"reference_labels_{participant}_{phase}_{channel}.csv"
            )
            reference_data = pd.read_csv(file_path)
            reference_data = reference_data.set_index(["heartbeat_id", "channel", "label"])

            reference_data_dict[phase] = reference_data

        if self.is_single(None):
            return reference_data_dict[phases[0]]
        return pd.concat(reference_data_dict, names=["phase"])

    @property
    def heartbeats(self) -> pd.DataFrame:
        """Segment heartbeats from the ECG data and return the heartbeat borders.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Heartbeats as a pandas DataFrame.

        """
        heartbeat_algo = HeartbeatSegmentationNeurokit()
        heartbeat_algo.extract(ecg=self.ecg, sampling_rate_hz=self.sampling_rate_ecg)
        heartbeats = heartbeat_algo.heartbeat_list_
        return heartbeats

    @staticmethod
    def _cut_to_labeling_borders(data: pd.DataFrame, borders: pd.DataFrame) -> pd.DataFrame:
        start = borders.index[0]
        end = borders.index[-1]
        data = data.loc[start:end]
        return data

    @property
    def metadata(self) -> pd.DataFrame:
        """Return metadata for the selected participants.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Metadata as a pandas DataFrame.

        """
        data = pd.read_csv(self.base_path.joinpath("metadata/demographics.csv"))
        data = data.set_index("participant")

        return data.loc[self.index["participant"].unique()]

    @property
    def age(self) -> pd.DataFrame:
        """Return the age of the selected participants.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Age as a pandas DataFrame.

        """
        return self.metadata[["Age"]]

    @property
    def gender(self) -> pd.DataFrame:
        """Return the gender of the selected participants.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Gender as a pandas DataFrame, recoded as {"M": "Male", "F": "Female"}

        """
        return self.metadata[["Gender"]].replace(self.GENDER_MAPPING)

    @property
    def bmi(self) -> pd.DataFrame:
        """Compute the BMI of the selected participants and return it.

        Returns
        -------
        :class:`~pandas.DataFrame`
            BMI as a pandas DataFrame.

        """
        return bmi(self.metadata[["Weight", "Height"]])



"""Some small utilities to improve writing docstrings.

For now, this just exports some functions from scipy._lib.doccer, to have only one place to import from.
While, the ``doccer`` submodule of scip[y is not part of the public API, it seems to be stable enough to use it here.
"""

from collections.abc import Callable

from scipy._lib.doccer import filldoc, inherit_docstring_from


def filldoc_with_better_error(docdict: dict[str, str]) -> filldoc:
    def inner(func: Callable) -> str:
        try:
            return filldoc(docdict)(func)
        except ValueError as e:
            if "unsupported format character" in str(e):
                raise ValueError(
                    "Your docstring contains a single '%' character. "
                    "This is not supported by this decorator. "
                    "If you want to type a '%' character, you need to type '%%'."
                ) from e
            raise

    return inner


def make_filldoc(docdict: dict[str, str], *, doc_summary: str | None = None) -> filldoc:
    """Create a new doc-filler from a dictionary.

    This can be applied to a function, method, or class to substitute ``%(key)s`` occurrences in its docstring.

    Parameters
    ----------
    docdict
        Dictionary with docstring keys and values.
        These can be multiline strings and will be dedented before substitution.
    doc_summary
        An optional summary line for the docstring of the returned decorator function.
        This is only relevant for inclusion in the documentation.

    Note
    ----
    If you use this decorator, your docstrings can not contain a single ``%`` character.
    If you want to type a ``%`` character, you need to type ``%%``.
    """
    inner = filldoc_with_better_error(docdict)

    if doc_summary is None:
        doc_summary = "Fill docstring from dictionary."

    full_doc = f"""{doc_summary}

    Available keys for the substitution::

    {list(docdict.keys())}

    """
    inner.__doc__ = full_doc
    inner._dict = docdict
    return inner


__all__ = ["inherit_docstring_from", "make_filldoc"]



import re

import numpy as np
import pandas as pd
from biopsykit.signals.ecg.segmentation import HeartbeatSegmentationNeurokit

from pepbench.utils._types import path_t


def _load_txt_data(file_path: path_t) -> pd.DataFrame:
    data = pd.read_csv(file_path, header=None)
    data.columns = ["icg", "icg_der", "ecg"]
    return data


def _get_match_heartbeat_label_ids(heartbeats: pd.DataFrame, b_points: pd.DataFrame) -> pd.Series:
    heartbeat_ids = pd.Series(index=heartbeats.index, name="heartbeat_id")
    heartbeat_ids.index.name = "heartbeat_id_b_point"
    for i, b_point in b_points.iterrows():
        idx = np.where(
            (heartbeats["start_sample"] <= b_point["sample_relative"])
            & (heartbeats["end_sample"] >= b_point["sample_relative"])
        )[0]
        if len(idx) == 0:
            # If no match is found, add NaN
            heartbeat_ids[i] = np.nan
        else:
            # If a match is found, add the index of the heartbeat
            heartbeat_ids[i] = idx[0]

    heartbeat_ids = heartbeat_ids.dropna().astype(int)
    return heartbeat_ids


def generate_heartbeat_borders(base_path: path_t) -> None:
    data_folder = base_path.joinpath("signals")
    annotation_folder = base_path.joinpath("annotations")
    heartbeat_folder = base_path.joinpath("reference_heartbeats")
    heartbeat_folder.mkdir(parents=True, exist_ok=True)
    for annotation_path in sorted(annotation_folder.glob("*.csv")):
        matches = re.findall(r"IDN(\d+).csv", str(annotation_path.name))
        p_id = matches[0]
        data_path = data_folder.joinpath(f"IDN{p_id}.txt")
        data = _load_txt_data(data_path)
        fs = 2000

        data.index /= fs
        data.index.name = "t"

        heartbeat_algo = HeartbeatSegmentationNeurokit()
        heartbeat_algo.extract(ecg=data[["ecg"]], sampling_rate_hz=fs)
        heartbeats = heartbeat_algo.heartbeat_list_

        heartbeats = heartbeats.round(2)

        heartbeat_path = heartbeat_folder.joinpath(f"IDN{p_id}.csv")
        heartbeats.to_csv(heartbeat_path)



"""Module to export results to LaTeX tables."""

from pepbench.export._latex import (
    convert_to_latex,
    create_algorithm_result_table,
    create_nan_reason_table,
    create_outlier_correction_table,
    create_reference_pep_table,
)

__all__ = [
    "convert_to_latex",
    "create_algorithm_result_table",
    "create_nan_reason_table",
    "create_outlier_correction_table",
    "create_reference_pep_table",
]



"""Module for plotting signals and results of PEP benchmarks."""

from pepbench.plotting import algorithms, results
from pepbench.plotting._base_plotting import (
    plot_signals,
    plot_signals_from_challenge_results,
    plot_signals_with_algorithm_results,
    plot_signals_with_reference_labels,
    plot_signals_with_reference_pep,
)

__all__ = [
    "algorithms",
    "plot_signals",
    "plot_signals_from_challenge_results",
    "plot_signals_with_algorithm_results",
    "plot_signals_with_reference_labels",
    "plot_signals_with_reference_pep",
    "results",
]



"""Module for loading and saving challenge results."""

from pepbench.io._io import convert_hz_to_ms, load_challenge_results_from_folder

__all__ = ["convert_hz_to_ms", "load_challenge_results_from_folder"]



"""Module providing custom styling functions for :class:`~pandas.DataFrame.style` objects."""

import pandas as pd

__all__ = ["highlight_min_per_group", "highlight_min_uncertainty", "highlight_outlier_improvement"]


def highlight_outlier_improvement(col: pd.Series) -> pd.Series:
    """Highlights the improvement of the outlier correction algorithm by changing the background color.

    This is a custom styling function that can be used with the :meth:`~pandas.DataFrame.style.apply` method.

    Parameters
    ----------
    col : :class:`~pandas.Series`
        A pandas Series object where each element belongs to a group identified by 'b_point_algorithm'.

    Returns
    -------
    :class:`~pandas.Series`
        A pandas Series object with the same index as the input col, where the element
        corresponding to the outlier improvement is styled with "background-color: Pink"
        and other elements are styled with "background-color: LightGreen".
    """
    if "b_point_algorithm" in col.index.names:
        idx_name = "b_point_algorithm"
    elif "B-Point Detection" in col.index.names:
        idx_name = "B-Point Detection"
    else:
        raise ValueError("Index name 'b_point_algorithm' or 'B-Point Detection' not found in the index names.")
    none_is_min = col.groupby(idx_name).transform(lambda s: any(t in s.idxmin() for t in ["none", "None"]))
    return none_is_min.map(
        {
            True: "background-color: Pink",
            False: "background-color: LightGreen",
        }
    )


def highlight_min_per_group(col: pd.Series) -> pd.Series:
    """Highlights the minimum value in each group by changing the background color.

    This is a custom styling function that can be used with the :meth:`~pandas.DataFrame.style.apply` method.

    Parameters
    ----------
    col : :class:`~pandas.Series`
        A pandas Series object where each element belongs to a group identified by 'b_point_algorithm'.

    Returns
    -------
    :class:`~pandas.Series`
        A pandas Series object with the same index as the input col, where the element
        corresponding to the minimum value in each group is styled with "background-color: LightGreen"
        and other elements are styled with an empty string.
    """
    idx_min = col.groupby("b_point_algorithm").idxmin()
    return (pd.Series(col.index.isin(idx_min), index=col.index)).map(
        {
            True: "background-color: LightGreen",
            False: "",
        }
    )


def highlight_min_uncertainty(row: pd.Series) -> pd.Series:
    """Highlights the minimum value of an uncertainty metric (mean ± std) in a row by making the font bold.

    This is a custom styling function that can be used with the :meth:`~pandas.DataFrame.style.apply` method.

    Parameters
    ----------
    row : :class:`~pandas.Series`
        A pandas Series object where each element is a string containing a numeric value
        followed by a space and additional text.

    Returns
    -------
    :class:`~pandas.Series`
        A pandas Series object with the same index as the input row, where the element
        corresponding to the minimum value is styled with "font-weight: bold;" and other
        elements are styled with an empty string.
    """
    row = row.apply(lambda s: float(s.split(" ")[0]))
    idx_min = row.index.isin([row.idxmin()])
    return (pd.Series(idx_min, index=row.index)).map(
        {
            True: "font-weight: bold;",
            False: "",
        }
    )



from collections.abc import Callable, Sequence

import numpy as np
from tpcp.validate import Aggregator


class PerSampleAggregator(Aggregator[np.ndarray]):
    def __init__(
        self,
        func: Callable[[Sequence[np.ndarray]], float | dict[str, float]],
        *,
        return_raw_scores: bool = True,
    ) -> None:
        self.func = func
        super().__init__(return_raw_scores=return_raw_scores)

    def aggregate(self, /, values: Sequence[np.ndarray], **_) -> dict[str, float]:  # noqa: ANN003
        return self.func(np.hstack(values))



import re
from collections.abc import Sequence
from functools import lru_cache
from pathlib import Path
from typing import ClassVar

import pandas as pd
from biopsykit.signals.ecg.event_extraction import QPeakExtractionVanLien2013
from biopsykit.signals.ecg.segmentation import HeartbeatSegmentationNeurokit
from biopsykit.signals.icg.preprocessing import IcgPreprocessingBandpass
from biopsykit.utils.dtypes import EcgRawDataFrame, IcgRawDataFrame

from pepbench.datasets import BasePepDatasetWithAnnotations
from pepbench.datasets._base_pep_extraction_dataset import base_pep_extraction_docfiller

__all__ = ["ReBeatIcgDataset"]

from pepbench.datasets.rebeat_icg._helper import _load_b_point_annotations, _load_mat_data
from pepbench.utils._types import path_t

_cached_get_mat_data = lru_cache(maxsize=10)(_load_mat_data)


@base_pep_extraction_docfiller
class ReBeatIcgDataset(BasePepDatasetWithAnnotations):
    base_path: Path
    use_cache: bool

    exclude_annotation_errors: bool

    SAMPLING_RATE: ClassVar[int] = 250

    SUBSET_ANNOTATION_ERRORS = (
        ("11", "Baseline"),
        ("11", "CognitiveWorkload"),
        ("12", "CognitiveWorkload"),
        ("16", "Baseline"),
        ("23", "Baseline"),
        ("23", "CognitiveWorkload"),
    )

    PHASES: ClassVar[dict[str, str]] = {"BL": "Baseline", "CW": "CognitiveWorkload"}
    PHASES_INVERSE: ClassVar[dict[str, str]] = {v: k for k, v in PHASES.items()}

    def __init__(
        self,
        base_path: path_t,
        groupby_cols: Sequence[str] | None = None,
        subset_index: Sequence[str] | None = None,
        *,
        return_clean: bool = True,
        exclude_annotation_errors: bool = True,
        use_cache: bool = True,
        only_labeled: bool = False,
    ) -> None:
        """Initialize a new ``GuardianDataset`` instance.

        Parameters
        ----------
        base_path : :class:`~pathlib.Path` or str
            Path to the root directory of the Guardian dataset.
        return_clean : bool
            Whether to return the preprocessed/cleaned ECG and ICG data when accessing the respective properties.
            Default: ``True``.
        use_cache : bool, optional
            Whether to use caching for loading TFM data. Default: ``True``.
        only_labeled : bool, optional
            Whether to only return segments that are labeled (i.e., cut the data to the labeling borders).
            This is necessary when using the dataset for evaluating the performance of PEP extraction algorithms or for
            training ML-based PEP extraction algorithms. Default: ``False``.
        """
        self.base_path = base_path
        self.use_cache = use_cache
        self.exclude_annotation_errors = exclude_annotation_errors
        self.data_to_exclude = self._find_data_to_exclude()
        super().__init__(
            groupby_cols=groupby_cols, subset_index=subset_index, return_clean=return_clean, only_labeled=only_labeled
        )

    def _sanitize_params(self) -> None:
        # ensure pathlib
        self.base_path = Path(self.base_path)

    def create_index(self) -> pd.DataFrame:
        self._sanitize_params()

        if self.only_labeled:
            label_border_folder = self.base_path.joinpath("04_LabelingAndHeartBeatBorders")
            file_list = sorted(label_border_folder.glob("*.csv"))
            label_region_ids = list(
                filter(
                    lambda x: len(x) > 0,
                    [
                        re.findall(r"LabelingBorders_Subject_(\d+)_task_(\w+)_label_(\d+).csv", path.name)
                        for path in file_list
                    ],
                )
            )
            matches = [label_region_id[0] for label_region_id in label_region_ids]
            matches = [(participant.zfill(2), phase, int(label_period)) for participant, phase, label_period in matches]
            index = pd.DataFrame(matches, columns=["participant", "phase", "label_period"])
            index = index.set_index(["participant", "phase"]).rename(index=self.PHASES, level="phase").sort_index()
        else:
            if self.return_clean:
                file_prefix = "FilteredData"
                data_path = self.base_path.joinpath(f"02_{file_prefix}")
            else:
                file_prefix = "RawData"
                data_path = self.base_path.joinpath(f"01_{file_prefix}")

            file_list = sorted(data_path.glob("*.mat"))
            matches = [re.findall(rf"{file_prefix}_Subject_(\d+)_task_(\w+).mat", path.name)[0] for path in file_list]
            # add leading zeros to participant numbers
            matches = [(participant.zfill(2), phase) for participant, phase in matches]
            index = pd.DataFrame(matches, columns=["participant", "phase"])
            index = index.set_index(["participant", "phase"]).rename(index=self.PHASES, level="phase").sort_index()

        index = index.reset_index()
        for item in self.data_to_exclude:
            index = index.drop(index[(index["participant"] == item[0]) & (index["phase"] == item[1])].index)
            index = index.reset_index(drop=True)
        return index

    def _find_data_to_exclude(self) -> Sequence[tuple[str, str]]:
        data_to_exclude = []
        if self.exclude_annotation_errors:
            data_to_exclude.extend(self.SUBSET_ANNOTATION_ERRORS)

        return data_to_exclude

    @property
    def sampling_rate_ecg(self) -> int:
        """Return the sampling rate of the ECG signal.

        Returns
        -------
        int
            Sampling rate of the ECG signal in Hz.

        """
        return self.SAMPLING_RATE

    @property
    def sampling_rate_icg(self) -> int:
        """Return the sampling rate of the ICG signal.

        Returns
        -------
        int
            Sampling rate of the ICG signal in Hz.

        """
        return self.SAMPLING_RATE

    @property
    def data(self) -> pd.DataFrame:
        if not self.is_single(None):
            if self.only_labeled:
                error_msg = "Data can only be accessed for a single participant, phase, and label period."
            else:
                error_msg = "Data can only be accessed for a single participant and phase."
            raise ValueError(error_msg)
        p_id = self.index["participant"][0]
        phase = self.index["phase"][0]

        if self.return_clean:
            file_prefix = "FilteredData"
            folder_path = self.base_path.joinpath(f"02_{file_prefix}")
        else:
            file_prefix = "RawData"
            folder_path = self.base_path.joinpath(f"01_{file_prefix}")

        file_path = folder_path.joinpath(f"{file_prefix}_Subject_{int(p_id)}_task_{self.PHASES_INVERSE[phase]}.mat")
        data = _cached_get_mat_data(file_path) if self.use_cache else _load_mat_data(file_path)

        if data["samplFreq"] != self.SAMPLING_RATE:
            raise ValueError(f"Sampling rate of {data['samplFreq']} does not match expected {self.SAMPLING_RATE}.")

        data = pd.DataFrame({"ecg": data["ECG"], "icg_der": data["ICG"]})
        data.index /= self.SAMPLING_RATE
        data.index.name = "t"
        data.index = pd.to_timedelta(data.index, unit="s")

        if self.only_labeled:
            labeling_borders = self.labeling_borders.iloc[0]
            data = data.iloc[labeling_borders["start_sample"] : labeling_borders["end_sample"]]
            data.index -= data.index[0]

        return data

    @property
    def ecg(self) -> EcgRawDataFrame:
        return self.data[["ecg"]]

    @property
    def icg(self) -> IcgRawDataFrame:
        icg = self.data[["icg_der"]]
        if self.return_clean:
            algo = IcgPreprocessingBandpass()
            algo.clean(icg=icg, sampling_rate_hz=self.sampling_rate_icg)
            icg = algo.icg_clean_
        return icg

    @property
    def heartbeats(self) -> pd.DataFrame:
        """Segment heartbeats from the ECG data and return the heartbeat borders.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Heartbeats as a pandas DataFrame.

        """
        heartbeat_algo = HeartbeatSegmentationNeurokit()
        heartbeat_algo.extract(ecg=self.ecg, sampling_rate_hz=self.sampling_rate_ecg)
        heartbeats = heartbeat_algo.heartbeat_list_
        return heartbeats

    @property
    def labeling_borders(self) -> pd.DataFrame:
        """Return the labeling borders for a selected participant and phase.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Labeling borders as :class:`~pandas.DataFrame`.

        """
        labeling_border_folder = self.base_path.joinpath("04_LabelingAndHeartBeatBorders")
        if not labeling_border_folder.exists():
            raise ValueError(
                "Labeling borders not found. Please generate them first by calling "
                "`generate_labeling_and_heartbeat_borders()`."
            )

        p_id = self.index["participant"][0]
        phase = self.index["phase"][0]
        label_period = int(self.index["label_period"][0])
        file_path = labeling_border_folder.joinpath(
            f"LabelingBorders_Subject_{int(p_id)}_task_{self.PHASES_INVERSE[phase]}_label_{label_period}.csv"
        )
        labeling_borders = pd.read_csv(file_path)
        labeling_borders = labeling_borders.astype(int)
        return labeling_borders

    @property
    def reference_heartbeats(self) -> pd.DataFrame:
        """Return the reference heartbeats.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference heartbeats as a pandas DataFrame

        """
        reference_heartbeat_folder = self.base_path.joinpath("04_LabelingAndHeartBeatBorders")
        if not reference_heartbeat_folder.exists():
            raise ValueError(
                "Reference heartbeats not found. Please generate them first by calling "
                "`pepbench.datasets.rebeat_icg.generate_labeling_and_heartbeat_borders()`."
            )

        p_id = self.index["participant"][0]
        phase = self.index["phase"][0]
        label_period = self.index["label_period"][0]
        file_path = reference_heartbeat_folder.joinpath(
            f"Heartbeats_Subject_{int(p_id)}_task_{self.PHASES_INVERSE[phase]}_label_{label_period}.csv"
        )
        heartbeats = pd.read_csv(file_path)
        heartbeats = heartbeats.set_index("heartbeat_id")
        heartbeats.index -= heartbeats.index[0]

        heartbeats = heartbeats.assign(
            start_sample=heartbeats["start_sample"] - self.labeling_borders["start_sample"].iloc[0],
            end_sample=heartbeats["end_sample"] - self.labeling_borders["start_sample"].iloc[0],
            r_peak_sample=heartbeats["r_peak_sample"] - self.labeling_borders["start_sample"].iloc[0],
        )

        return heartbeats

    @property
    def reference_labels_icg(self) -> pd.DataFrame:
        """Return the reference labels for the ICG data.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference labels for the ICG data as a pandas DataFrame

        """
        heartbeats = self.reference_heartbeats[["start_sample", "end_sample"]]
        b_points = self._load_b_point_annotations()
        labeling_borders = self.labeling_borders.iloc[0]

        b_points = b_points[
            (b_points["sample_relative"] > labeling_borders["start_sample"])
            & (b_points["sample_relative"] < labeling_borders["end_sample"])
        ]
        b_points -= labeling_borders["start_sample"]
        b_points.index -= b_points.index[0]

        data = pd.concat(
            {
                ("heartbeats", "start"): heartbeats["start_sample"],
                ("heartbeats", "end"): heartbeats["end_sample"],
                ("ICG", "B-point"): b_points["sample_relative"],
            },
            axis=1,
        )
        data = data.stack([0, 1], future_stack=True).astype("Int64")
        data = pd.DataFrame(data, columns=["sample_relative"])
        data.index = data.index.set_names(["heartbeat_id", "channel", "label"])
        data = data.sort_values(by="sample_relative")
        return data

    @property
    def reference_labels_ecg(self) -> pd.DataFrame:
        """Return the reference labels for the ECG data.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference labels for the ECG data as a pandas DataFrame

        """
        ecg = self.ecg
        q_peak_algo = QPeakExtractionVanLien2013(time_interval_ms=32)
        q_peak_algo.extract(ecg=ecg, heartbeats=self.reference_heartbeats, sampling_rate_hz=self.sampling_rate_ecg)
        q_peaks = q_peak_algo.points_
        q_peaks = q_peaks[["q_peak_sample"]]
        heartbeats = self.reference_heartbeats[["start_sample", "end_sample"]]
        heartbeats.columns = ["start", "end"]
        q_peaks.columns = ["Q-peak"]

        res = pd.concat({"heartbeat": heartbeats, "ECG": q_peaks}, axis=1)
        res = res.stack([0, 1], future_stack=True).sort_values().to_frame(name="sample_relative")
        res.index = res.index.set_names(["heartbeat_id", "channel", "label"])

        return res

    def _load_b_point_annotations(self) -> pd.DataFrame:
        p_id = self.index["participant"][0]
        phase = self.index["phase"][0]
        file_path = self.base_path.joinpath(
            f"03_ExpertAnnotations/Annotat_Subject_{int(p_id)}_task_{self.PHASES_INVERSE[phase]}.mat"
        )
        return _load_b_point_annotations(file_path, load_func=_cached_get_mat_data if self.use_cache else None)



"""Module for various data handling helper functions."""

from pepbench.data_handling import utils
from pepbench.data_handling._data_handling import (
    add_unique_id_to_results_dataframe,
    compute_improvement_outlier_correction,
    compute_improvement_pipeline,
    compute_pep_performance_metrics,
    correlation_reference_pep_heart_rate,
    describe_pep_values,
    get_data_for_algo,
    get_error_by_group,
    get_pep_for_algo,
    get_reference_data,
    get_reference_pep,
    rr_interval_to_heart_rate,
)

__all__ = [
    "add_unique_id_to_results_dataframe",
    "compute_improvement_outlier_correction",
    "compute_improvement_pipeline",
    "compute_pep_performance_metrics",
    "correlation_reference_pep_heart_rate",
    "describe_pep_values",
    "get_data_for_algo",
    "get_error_by_group",
    "get_pep_for_algo",
    "get_reference_data",
    "get_reference_pep",
    "rr_interval_to_heart_rate",
    "utils",
]



import pandas as pd

from pepbench.datasets.guardian._tfm_loader import TFMLoader
from pepbench.utils._types import path_t


def _load_tfm_data(base_path: path_t, date: pd.Timestamp) -> dict[str, pd.DataFrame]:
    tfm_data = TFMLoader.from_mat_file(base_path, date)
    return tfm_data.data_as_dict(index="local_datetime")



import re
import warnings
from collections.abc import Callable

import numpy as np
import pandas as pd
from biopsykit.signals.ecg.segmentation import HeartbeatSegmentationNeurokit
from scipy.io import loadmat

from pepbench.utils._types import path_t


def _load_mat_data(file_path: path_t) -> dict[str, np.ndarray]:
    return loadmat(file_path, squeeze_me=True)


def generate_labeling_and_heartbeat_borders(base_path: path_t) -> None:
    data_folder = base_path.joinpath("01_RawData")
    annotation_folder = base_path.joinpath("03_ExpertAnnotations")
    labeling_border_folder = base_path.joinpath("04_LabelingAndHeartBeatBorders")
    labeling_border_folder.mkdir(parents=True, exist_ok=True)
    for annotation_path in sorted(annotation_folder.glob("*.mat")):
        matches = re.findall(r"Annotat_Subject_(\d+)_task_(\w+).mat", str(annotation_path.name))
        p_id = matches[0][0]
        phase = matches[0][1]
        b_points = _load_b_point_annotations(annotation_path)

        data_path = data_folder.joinpath(f"RawData_Subject_{int(p_id)}_task_{phase}.mat")
        data = _load_mat_data(data_path)
        fs = data["samplFreq"]

        data = pd.DataFrame({"ecg": data["ECG"], "icg_der": data["ICG"]})
        data.index /= fs
        data.index.name = "t"

        heartbeat_algo = HeartbeatSegmentationNeurokit()
        heartbeat_algo.extract(ecg=data[["ecg"]], sampling_rate_hz=fs)
        heartbeats = heartbeat_algo.heartbeat_list_

        warnings.simplefilter("ignore", FutureWarning)
        b_point_groups = np.split(b_points, np.where(b_points.diff() > 5 * fs)[0])
        b_point_groups = pd.concat(dict(enumerate(b_point_groups)))
        b_point_groups.index = b_point_groups.index.set_names("label_region", level=0)
        b_point_groups = b_point_groups.reset_index()

        heartbeat_ids = []
        for _i, b_point in b_point_groups.iterrows():
            idx = np.where(
                (heartbeats["start_sample"] <= b_point["sample_relative"])
                & (heartbeats["end_sample"] >= b_point["sample_relative"])
            )[0]
            if len(idx) != 0:
                heartbeat_ids.append(idx[0])
        heartbeat_ids = pd.Series(heartbeat_ids)

        heartbeats = heartbeats.loc[heartbeat_ids]
        heartbeats.index = b_point_groups.index
        heartbeats.index -= heartbeats.index[0]
        heartbeats = heartbeats.assign(label_region=b_point_groups["label_region"].astype(int))
        heartbeats.index.name = "heartbeat_id"

        labeling_borders = heartbeats.groupby("label_region").apply(
            lambda s, sampling_rate=fs: {
                "start_sample": s.iloc[0]["start_sample"] - int(0.5 * sampling_rate),
                "end_sample": s.iloc[-1]["end_sample"] + int(0.5 * sampling_rate),
            },
        )
        labeling_borders = labeling_borders.apply(pd.Series).astype("Int64")

        for label_id, df in heartbeats.groupby("label_region"):
            df_out = df.drop(columns=["label_region"])
            heartbeat_path = labeling_border_folder.joinpath(
                f"Heartbeats_Subject_{int(p_id)}_task_{phase}_label_{label_id}.csv"
            )
            df_out = df_out.astype(
                {
                    "start_sample": "Int64",
                    "end_sample": "Int64",
                    "r_peak_sample": "Int64",
                    "rr_interval_sample": "Int64",
                }
            )
            df_out.to_csv(heartbeat_path)
            labeling_borders_path = labeling_border_folder.joinpath(
                f"LabelingBorders_Subject_{int(p_id)}_task_{phase}_label_{label_id}.csv"
            )
            labeling_borders.iloc[[label_id]].to_csv(labeling_borders_path, index=False)


def _load_b_point_annotations(file_path: path_t, load_func: Callable | None = None) -> pd.DataFrame:
    if load_func is None:
        load_func = _load_mat_data
    annotations = load_func(file_path)
    annotations = pd.DataFrame(annotations["annotPoints"], columns=["B", "C", "X", "Other"])
    annotations = annotations[["B"]].sort_values(by="B").reset_index(drop=True)
    annotations.columns = ["sample_relative"]
    annotations.index.name = "heartbeat_id"
    return annotations
